from mage_ai.api.errors import ApiError


class BaseMonitor():
    def __init__(self, resource, user, error, **kwargs):
        self.error = error
        self.options = kwargs
        self.resource = resource
        self.user = user

    def present(self):
        data = ApiError.RESOURCE_ERROR.copy()
        if self.error.code:
            data['code'] = self.error.code
        if self.error.errors:
            data['errors'] = self.error.errors
        if self.error.message:
            data['message'] = self.error.message
        if self.error.type:
            data['type'] = self.error.type
        return data



from contextlib import redirect_stdout
from datetime import datetime
from jinja2 import Template
from logging import Logger
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.sql import (
    bigquery,
    execute_sql_code as execute_sql_code_orig,
    mssql,
    mysql,
    postgres,
    redshift,
    snowflake,
    spark,
    trino,
)
from mage_ai.data_preparation.models.constants import BlockLanguage, BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.stream import StreamToLogger
from mage_ai.data_preparation.shared.utils import get_template_vars
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.io.base import DataSource, ExportWritePolicy
from mage_ai.io.config import ConfigFileLoader
from mage_ai.orchestration.constants import PIPELINE_RUN_MAGE_VARIABLES_KEY
from mage_ai.shared.array import find
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.strings import remove_extension_from_filename
from mage_ai.shared.utils import clean_name, files_in_path
from pandas import DataFrame
from typing import Callable, Dict, List, Tuple
import aiofiles
import os
import re
import shutil
import simplejson
import subprocess
import sys
import uuid
import yaml


PROFILES_FILE_NAME = 'profiles.yml'


def parse_attributes(block) -> Dict:
    configuration = block.configuration

    file_path = configuration['file_path']
    path_parts = file_path.split(os.sep)
    project_name = path_parts[0]
    filename = path_parts[-1]
    model_name = None
    file_extension = None

    parts = filename.split('.')
    if len(parts) >= 2:
        model_name = '.'.join(parts[:-1])
        file_extension = parts[-1]

    # Check the model SQL file content for a config with an alias value. If it exists,
    # use that alias value as the table name instead of the model’s name.
    table_name = model_name
    config = model_config(block.content)
    if config.get('alias'):
        table_name = config['alias']

    full_path = os.path.join(get_repo_path(), 'dbt', file_path)

    project_full_path = os.path.join(get_repo_path(), 'dbt', project_name)

    models_folder_path = os.path.join(project_full_path, 'models')
    sources_full_path = os.path.join(models_folder_path, 'mage_sources.yml')
    sources_full_path_legacy = full_path.replace(filename, 'mage_sources.yml')

    profiles_full_path = os.path.join(project_full_path, PROFILES_FILE_NAME)
    profile_target = configuration.get('dbt_profile_target')
    profile = load_profile(project_name, profiles_full_path, profile_target)

    dbt_project_full_path = os.path.join(project_full_path, 'dbt_project.yml')

    source_name = f'mage_{project_name}'
    if profile:
        if (DataSource.MYSQL == profile.get('type') or
                DataSource.REDSHIFT == profile.get('type') or
                DataSource.TRINO == profile.get('type') or
                DataSource.MSSQL == profile.get('type') or
                DataSource.SPARK == profile.get('type')):
            source_name = profile['schema']

    return dict(
        dbt_project_full_path=dbt_project_full_path,
        file_extension=file_extension,
        file_path=file_path,
        filename=filename,
        full_path=full_path,
        model_name=model_name,
        models_folder_path=models_folder_path,
        profiles_full_path=profiles_full_path,
        project_full_path=project_full_path,
        project_name=project_name,
        source_name=source_name,
        sources_full_path=sources_full_path,
        sources_full_path_legacy=sources_full_path_legacy,
        table_name=table_name,
    )


def extract_refs(block_content) -> List[str]:
    return re.findall(
        r"{}[ ]*ref\(['\"]+([\w]+)['\"]+\)[ ]*{}".format(r'\{\{', r'\}\}'),
        block_content,
    )


def extract_sources(block_content) -> List[Tuple[str, str]]:
    return re.findall(
        r"{}[ ]*source\(['\"]+([\w]+)['\"]+[,]+[ ]*['\"]+([\w]+)['\"]+\)[ ]*{}".format(
            r'\{\{',
            r'\}\}',
        ),
        block_content,
    )


def add_blocks_upstream_from_refs(
    block: 'Block',
    add_current_block: bool = False,
    downstream_blocks: List['Block'] = [],
    read_only: bool = False,
) -> None:
    attributes_dict = parse_attributes(block)
    models_folder_path = attributes_dict['models_folder_path']

    files_by_name = {}
    for file_path_orig in files_in_path(models_folder_path):
        file_path = file_path_orig.replace(f'{models_folder_path}{os.sep}', '')
        filename = file_path.split(os.sep)[-1]
        parts = filename.split('.')
        if len(parts) >= 2:
            fn = '.'.join(parts[:-1])
            file_extension = parts[-1]
            if 'sql' == file_extension:
                files_by_name[fn] = file_path_orig

    current_upstream_blocks = []
    added_blocks = []
    for idx, ref in enumerate(extract_refs(block.content)):
        if ref not in files_by_name:
            print(f'WARNING: dbt model {ref} cannot be found.')
            continue

        fp = files_by_name[ref].replace(f"{os.path.join(get_repo_path(), 'dbt')}{os.sep}", '')
        configuration = dict(file_path=fp)
        uuid = remove_extension_from_filename(fp)

        if read_only:
            uuid_clean = clean_name(uuid, allow_characters=[os.sep])
            new_block = block.__class__(uuid_clean, uuid_clean, block.type)
            new_block.configuration = configuration
            new_block.language = block.language
            new_block.pipeline = block.pipeline
            new_block.downstream_blocks = [block]
            new_block.upstream_blocks = add_blocks_upstream_from_refs(
                new_block,
                read_only=read_only,
            )
            added_blocks += new_block.upstream_blocks
        else:
            new_block = block.__class__.create(
                uuid,
                block.type,
                get_repo_path(),
                configuration=configuration,
                language=block.language,
                pipeline=block.pipeline,
            )

        added_blocks.append(new_block)
        current_upstream_blocks.append(new_block)

    if add_current_block:
        arr = []
        for b in current_upstream_blocks:
            arr.append(b)
        block.upstream_blocks = arr
        added_blocks.append(block)

    return added_blocks


def get_source(block) -> Dict:
    attributes_dict = parse_attributes(block)
    source_name = attributes_dict['source_name']
    settings = load_sources(block)
    return find(lambda x: x['name'] == source_name, settings.get('sources', []))


def load_sources(block) -> Dict:
    attributes_dict = parse_attributes(block)
    sources_full_path = attributes_dict['sources_full_path']
    sources_full_path_legacy = attributes_dict['sources_full_path_legacy']

    settings = None
    if os.path.exists(sources_full_path):
        with open(sources_full_path, 'r') as f:
            settings = yaml.safe_load(f) or dict(sources=[], version=2)

    if os.path.exists(sources_full_path_legacy):
        print(f'Legacy dbt source file exists at {sources_full_path_legacy}.')

        with open(sources_full_path_legacy, 'r') as f:
            sources_legacy = yaml.safe_load(f) or dict(sources=[], version=2)

            for source_data in sources_legacy.get('sources', []):
                source_name = source_data['name']
                for table_data in source_data['tables']:
                    table_name = table_data['name']
                    print(f'Adding source {source_name} and table {table_name} '
                          f'to {sources_full_path}.')
                    settings = add_table_to_source(block, settings, source_name, table_name)

            with open(sources_full_path_legacy, 'w') as f:
                print(f'Deleting legacy dbt source file at {sources_full_path_legacy}.')
                yaml.safe_dump(settings, f)

        os.remove(sources_full_path_legacy)

    return settings


def source_table_name_for_block(block) -> str:
    return f'{clean_name(block.pipeline.uuid)}_{clean_name(block.uuid)}'


def update_model_settings(
    block: 'Block',
    upstream_blocks: List['Block'],
    upstream_blocks_previous: List['Block'],
    force_update: bool = False,
):
    attributes_dict = parse_attributes(block)

    sources_full_path = attributes_dict['sources_full_path']
    source_name = attributes_dict['source_name']

    if not force_update and len(upstream_blocks_previous) > len(upstream_blocks):
        # TODO (tommy dangerous): should we remove sources?
        # How do we know no other model is using a source?

        # uuids = [b.uuid for b in upstream_blocks]
        # for upstream_block in upstream_blocks_previous:
        #     if upstream_block.uuid in uuids:
        #         continue

        #     # If upstream block that’s being removed has a downstream block that is a DBT block
        #     if any([block.type == b.type for b in upstream_block.downstream_blocks]):
        #         continue

        #     if os.path.exists(sources_full_path):
        #         with open(sources_full_path, 'r') as f:
        #             settings = yaml.safe_load(f) or {}
        #             source = find(lambda x: x['name'] == source_name, settings.get('sources', []))
        #             table_name = f'{upstream_block.pipeline.uuid}_{upstream_block.uuid}'
        #             if source:
        #                 source['tables'] = list(
        #                     filter(
        #                         lambda x: x['name'] != table_name,
        #                         source.get('tables', []),
        #                     ),
        #                 )

        #         with open(sources_full_path, 'w') as f:
        #             yaml.safe_dump(settings, f)
        pass
    elif upstream_blocks:
        for upstream_block in upstream_blocks:
            if block.type == upstream_block.type:
                continue

            table_name = source_table_name_for_block(upstream_block)
            settings = add_table_to_source(block, load_sources(block), source_name, table_name)

            with open(sources_full_path, 'w') as f:
                yaml.safe_dump(settings, f)


def add_table_to_source(block: 'Block', settings: Dict, source_name: str, table_name: str) -> None:
    new_table = dict(name=table_name)
    new_source = dict(
        name=source_name,
        tables=[
            new_table,
        ],
    )

    if settings:
        source = find(lambda x: x['name'] == source_name, settings.get('sources', []))
        if source:
            if not source.get('tables'):
                source['tables'] = []
            if table_name not in [x['name'] for x in source['tables']]:
                source['tables'].append(new_table)
        else:
            settings['sources'].append(new_source)

    else:
        settings = dict(
            version=2,
            sources=[
                new_source,
            ],
        )

    return settings


def load_profiles_file(profiles_full_path: str) -> Dict:
    try:
        with open(profiles_full_path, 'r') as f:
            try:
                text = Template(f.read()).render(
                    **get_template_vars(),
                )
                return yaml.safe_load(text)
            except Exception as err:
                print(
                    f'Error loading file {profiles_full_path}, check file content syntax: {err}.',
                )
                return {}
    except OSError as err:
        print(
            f'Error loading file {profiles_full_path}, check file content syntax: {err}.',
        )
        return {}


async def load_profiles_file_async(profiles_full_path: str) -> Dict:
    try:
        async with aiofiles.open(profiles_full_path, mode='r') as fp:
            try:
                file_content = await fp.read()
                text = Template(file_content).render(
                    **get_template_vars(),
                )
                return yaml.safe_load(text)
            except Exception as err:
                print(
                    f'Error loading file {profiles_full_path}, check file content syntax: {err}.',
                )
                return {}
    except OSError as err:
        print(
            f'Error loading file {profiles_full_path}, check file content syntax: {err}.',
        )
        return {}


def load_profiles(project_name: str, profiles_full_path: str) -> Dict:
    profiles = load_profiles_file(profiles_full_path)

    if not profiles or project_name not in profiles:
        print(f'Project name {project_name} does not exist in profile file {profiles_full_path}.')
        return {}

    return profiles[project_name]


async def load_profiles_async(project_name: str, profiles_full_path: str) -> Dict:
    profiles = await load_profiles_file_async(profiles_full_path)

    if not profiles or project_name not in profiles:
        print(f'Project name {project_name} does not exist in profile file {profiles_full_path}.')
        return {}

    return profiles[project_name]


def load_profile(
    project_name: str,
    profiles_full_path: str,
    profile_target: str = None,
) -> Dict:

    profile = load_profiles(project_name, profiles_full_path)
    outputs = profile.get('outputs', {})
    target = profile.get('target', None)

    return outputs.get(profile_target or target)


def get_profile(block, profile_target: str = None) -> Dict:
    attr = parse_attributes(block)
    project_name = attr['project_name']
    profiles_full_path = attr['profiles_full_path']

    return load_profile(project_name, profiles_full_path, profile_target)


def config_file_loader_and_configuration(block, profile_target: str) -> Dict:
    profile = get_profile(block, profile_target)

    if not profile:
        raise Exception(
            f'No profile target named {profile_target}, check the {PROFILES_FILE_NAME} file.',
        )
    profile_type = profile.get('type')

    config_file_loader = None
    configuration = None

    if DataSource.POSTGRES == profile_type:
        database = profile.get('dbname')
        host = profile.get('host')
        password = profile.get('password')
        port = profile.get('port')
        schema = profile.get('schema')
        user = profile.get('user')

        config_file_loader = ConfigFileLoader(config=dict(
            POSTGRES_DBNAME=database,
            POSTGRES_HOST=host,
            POSTGRES_PASSWORD=password,
            POSTGRES_PORT=port,
            POSTGRES_SCHEMA=schema,
            POSTGRES_USER=user,
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=database,
            data_provider_schema=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.BIGQUERY == profile_type:
        keyfile = profile.get('keyfile')
        database = profile.get('project')
        schema = profile.get('dataset')

        config_file_loader = ConfigFileLoader(config=dict(
            GOOGLE_SERVICE_ACC_KEY_FILEPATH=keyfile,
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=database,
            data_provider_schema=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.MSSQL == profile_type:
        config_file_loader = ConfigFileLoader(config=dict(
            MSSQL_DATABASE=profile.get('database'),
            MSSQL_DRIVER=profile.get('driver'),
            MSSQL_HOST=profile.get('server'),
            MSSQL_PASSWORD=profile.get('password'),
            MSSQL_PORT=profile.get('port'),
            MSSQL_USER=profile.get('user'),
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=profile.get('database'),
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.MYSQL == profile_type:
        host = profile.get('server')
        password = profile.get('password')
        port = profile.get('port')
        schema = profile.get('schema')
        ssl_disabled = profile.get('ssl_disabled')
        username = profile.get('username')

        config_file_loader = ConfigFileLoader(config=dict(
            MYSQL_CONNECTION_METHOD='ssh_tunnel' if not ssl_disabled else None,
            MYSQL_DATABASE=schema,
            MYSQL_HOST=host,
            MYSQL_PASSWORD=password,
            MYSQL_PORT=port,
            MYSQL_USER=username,
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.REDSHIFT == profile_type:
        database = profile.get('dbname')
        host = profile.get('host')
        password = profile.get('password')
        port = profile.get('port', 5439)
        schema = profile.get('schema')
        user = profile.get('user')

        config_file_loader = ConfigFileLoader(config=dict(
            REDSHIFT_DBNAME=database,
            REDSHIFT_HOST=host,
            REDSHIFT_PORT=port,
            REDSHIFT_SCHEMA=schema,
            REDSHIFT_TEMP_CRED_PASSWORD=password,
            REDSHIFT_TEMP_CRED_USER=user,
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=database,
            data_provider_schema=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.SNOWFLAKE == profile_type:
        database = profile.get('database')
        schema = profile.get('schema')
        config = dict(
            SNOWFLAKE_ACCOUNT=profile.get('account'),
            SNOWFLAKE_DEFAULT_DB=profile.get('database'),
            SNOWFLAKE_DEFAULT_SCHEMA=profile.get('schema'),
            SNOWFLAKE_DEFAULT_WH=profile.get('warehouse'),
            SNOWFLAKE_USER=profile.get('user'),
        )

        if 'password' in profile:
            config['SNOWFLAKE_PASSWORD'] = profile['password']
        if 'private_key_passphrase' in profile:
            config['SNOWFLAKE_PRIVATE_KEY_PASSPHRASE'] = profile['private_key_passphrase']
        if 'private_key_path' in profile:
            config['SNOWFLAKE_PRIVATE_KEY_PATH'] = profile['private_key_path']

        config_file_loader = ConfigFileLoader(config=config)
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=database,
            data_provider_schema=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.SPARK == profile_type:
        schema = profile.get('schema')

        config = dict(
            SPARK_METHOD=profile.get('method'),
            SPARK_HOST=profile.get('host'),
            SPARK_SCHEMA=profile.get('schema'),
        )

        config_file_loader = ConfigFileLoader(config=config)
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )
    elif DataSource.TRINO == profile_type:
        catalog = profile.get('database')
        schema = profile.get('schema')

        config_file_loader = ConfigFileLoader(config=dict(
            TRINO_CATALOG=catalog,
            TRINO_HOST=profile.get('host'),
            TRINO_PASSWORD=profile.get('password'),
            TRINO_PORT=profile.get('port'),
            TRINO_SCHEMA=schema,
            TRINO_USER=profile.get('user'),
        ))
        configuration = dict(
            data_provider=profile_type,
            data_provider_database=catalog,
            data_provider_schema=schema,
            export_write_policy=ExportWritePolicy.REPLACE,
        )

    if not config_file_loader or not configuration:
        attr = parse_attributes(block)
        profiles_full_path = attr['profiles_full_path']

        msg = (f'No configuration matching profile type {profile_type}. '
               f'Change your target in {profiles_full_path} '
               'or add dbt_profile_target to your global variables.')
        raise Exception(msg)

    return config_file_loader, configuration


def execute_sql_code(
    block,
    query: str,
    profile_target: str,
    **kwargs,
):
    config_file_loader, configuration = config_file_loader_and_configuration(
        block,
        profile_target,
    )

    return execute_sql_code_orig(
        block,
        query,
        config_file_loader=config_file_loader,
        configuration=configuration,
        **kwargs,
    )


def create_upstream_tables(
    block,
    profile_target: str,
    cache_upstream_dbt_models: bool = False,
    **kwargs,
) -> None:
    if len([b for b in block.upstream_blocks if BlockType.SENSOR != b.type]) == 0:
        return

    config_file_loader, configuration = config_file_loader_and_configuration(
        block,
        profile_target,
    )

    data_provider = configuration.get('data_provider')

    kwargs_shared = merge_dict(dict(
        configuration=configuration,
        cache_upstream_dbt_models=cache_upstream_dbt_models,
    ), kwargs)

    upstream_blocks_init = block.upstream_blocks
    upstream_blocks = upstream_blocks_from_sources(block)
    block.upstream_blocks = upstream_blocks

    if DataSource.POSTGRES == data_provider:
        from mage_ai.io.postgres import Postgres

        with Postgres.with_config(config_file_loader) as loader:
            postgres.create_upstream_block_tables(
                loader,
                block,
                cascade_on_drop=True,
                **kwargs_shared,
            )
    elif DataSource.MSSQL == data_provider:
        from mage_ai.io.mssql import MSSQL

        with MSSQL.with_config(config_file_loader) as loader:
            mssql.create_upstream_block_tables(
                loader,
                block,
                cascade_on_drop=False,
                **kwargs_shared,
            )
    elif DataSource.MYSQL == data_provider:
        from mage_ai.io.mysql import MySQL

        with MySQL.with_config(config_file_loader) as loader:
            mysql.create_upstream_block_tables(
                loader,
                block,
                cascade_on_drop=True,
                **kwargs_shared,
            )
    elif DataSource.BIGQUERY == data_provider:
        from mage_ai.io.bigquery import BigQuery

        loader = BigQuery.with_config(config_file_loader)
        bigquery.create_upstream_block_tables(
            loader,
            block,
            configuration=configuration,
            cache_upstream_dbt_models=cache_upstream_dbt_models,
            **kwargs,
        )
    elif DataSource.REDSHIFT == data_provider:
        from mage_ai.io.redshift import Redshift

        with Redshift.with_config(config_file_loader) as loader:
            redshift.create_upstream_block_tables(
                loader,
                block,
                cascade_on_drop=True,
                **kwargs_shared,
            )
    elif DataSource.SNOWFLAKE == data_provider:
        from mage_ai.io.snowflake import Snowflake

        with Snowflake.with_config(config_file_loader) as loader:
            snowflake.create_upstream_block_tables(
                loader,
                block,
                **kwargs_shared,
            )
    elif DataSource.SPARK == data_provider:
        from mage_ai.io.spark import Spark

        loader = Spark.with_config(config_file_loader)
        spark.create_upstream_block_tables(
            loader,
            block,
            **kwargs_shared,
        )
    elif DataSource.TRINO == data_provider:
        from mage_ai.io.trino import Trino

        with Trino.with_config(config_file_loader) as loader:
            trino.create_upstream_block_tables(
                loader,
                block,
                **kwargs_shared,
            )

    block.upstream_blocks = upstream_blocks_init


def interpolate_input(
    block,
    query: str,
    configuration: Dict,
    profile_database: str,
    profile_schema: str,
    quote_str: str = '',
    replace_func=None,
) -> str:
    def __quoted(name):
        return quote_str + name + quote_str

    def __replace_func(db, schema, tn):
        if replace_func:
            return replace_func(db, schema, tn)

        if db and not schema:
            return f'{__quoted(db)}.{__quoted(tn)}'

        return f'{__quoted(schema)}.{__quoted(tn)}'

    for idx, upstream_block in enumerate(block.upstream_blocks):
        if BlockType.DBT != upstream_block.type:
            continue

        attrs = parse_attributes(upstream_block)
        table_name = attrs['table_name']

        arr = []
        if profile_database:
            arr.append(__quoted(profile_database))
        if profile_schema:
            arr.append(__quoted(profile_schema))
        if table_name:
            arr.append(__quoted(table_name))
        matcher1 = '.'.join(arr)

        database = configuration.get('data_provider_database')
        schema = configuration.get('data_provider_schema')
        table_name = upstream_block.table_name

        query = query.replace(
            matcher1,
            __replace_func(database, schema, table_name),
        )

    return query


def interpolate_refs_with_table_names(
    query_string: str,
    block: Block,
    profile_target: str,
    configuration: Dict,
):
    profile = get_profile(block, profile_target)

    profile_type = profile.get('type')
    quote_str = ''
    if DataSource.POSTGRES == profile_type:
        database = profile['dbname']
        schema = profile['schema']
        quote_str = '"'
    elif DataSource.MSSQL == profile_type:
        database = configuration['data_provider_database']
        schema = None
        quote_str = '`'
    elif DataSource.MYSQL == profile_type:
        database = configuration['data_provider_database']
        schema = None
        quote_str = '`'
    elif DataSource.BIGQUERY == profile_type:
        database = profile['project']
        schema = profile['dataset']
        quote_str = '`'
    elif DataSource.REDSHIFT == profile_type:
        database = profile['dbname']
        schema = profile['schema']
        quote_str = '"'
    elif DataSource.SNOWFLAKE == profile_type:
        database = profile['database']
        schema = profile['schema']
    elif DataSource.SPARK == profile_type:
        database = profile['schema']
        schema = None
    elif DataSource.TRINO == profile_type:
        database = profile['catalog']
        schema = profile['schema']

    return interpolate_input(
        block,
        query_string,
        configuration=configuration,
        profile_database=database,
        profile_schema=schema,
        quote_str=quote_str,
    )


def compiled_query_string(block: Block) -> str:
    attr = parse_attributes(block)

    project_full_path = attr['project_full_path']
    file_path = attr['file_path']

    file = os.path.join(project_full_path, 'target', 'compiled', file_path)

    if not os.path.exists(file):
        return None

    with open(file, 'r') as f:
        query_string = f.read()

        # TODO (tommy dang): this was needed because we didn’t want to create model tables and
        # so we’d create a table to store the model results without creating the model.
        # However, we’re requiring people to run the model and create the model table to use ref.
        # query_string = interpolate_refs_with_table_names(
        #     query_string,
        #     block,
        #     profile_target=profile_target,
        #     configuration=configuration,
        # )

    return query_string


def execute_query(
    block,
    profile_target: str,
    query_string: str,
    limit: int = None,
) -> DataFrame:
    config_file_loader, configuration = config_file_loader_and_configuration(
        block,
        profile_target,
    )

    data_provider = configuration['data_provider']

    shared_kwargs = {}
    if limit is not None:
        shared_kwargs['limit'] = limit

    if DataSource.POSTGRES == data_provider:
        from mage_ai.io.postgres import Postgres

        with Postgres.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)
    elif DataSource.MSSQL == data_provider:
        from mage_ai.io.mssql import MSSQL

        with MSSQL.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)
    elif DataSource.MYSQL == data_provider:
        from mage_ai.io.mysql import MySQL

        with MySQL.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)
    elif DataSource.BIGQUERY == data_provider:
        from mage_ai.io.bigquery import BigQuery

        loader = BigQuery.with_config(config_file_loader)
        return loader.load(query_string, **shared_kwargs)
    elif DataSource.REDSHIFT == data_provider:
        from mage_ai.io.redshift import Redshift

        with Redshift.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)
    elif DataSource.SNOWFLAKE == data_provider:
        from mage_ai.io.snowflake import Snowflake

        with Snowflake.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)
    elif DataSource.SPARK == data_provider:
        from mage_ai.io.spark import Spark

        loader = Spark.with_config(config_file_loader)
        return loader.load(query_string, **shared_kwargs)
    elif DataSource.TRINO == data_provider:
        from mage_ai.io.trino import Trino

        with Trino.with_config(config_file_loader) as loader:
            return loader.load(query_string, **shared_kwargs)


def query_from_compiled_sql(block, profile_target: str, limit: int = None) -> DataFrame:
    query_string = compiled_query_string(block)

    return execute_query(block, profile_target, query_string, limit)


def build_command_line_arguments(
    block,
    variables: Dict,
    run_settings: Dict = None,
    run_tests: bool = False,
    test_execution: bool = False,
) -> Tuple[str, List[str], Dict]:
    variables = merge_dict(
        variables or {},
        get_global_variables(block.pipeline.uuid) if block.pipeline else {},
    )
    dbt_command = 'run'

    if run_tests:
        dbt_command = 'test'

    if run_settings:
        if run_settings.get('build_model'):
            dbt_command = 'build'
        elif run_settings.get('test_model'):
            dbt_command = 'test'

    variables_json = {}
    for k, v in variables.items():
        if PIPELINE_RUN_MAGE_VARIABLES_KEY == k:
            continue

        if (type(v) is str or
                type(v) is int or
                type(v) is bool or
                type(v) is float or
                type(v) is dict or
                type(v) is list or
                type(v) is datetime):
            variables_json[k] = v

    args = [
        '--vars',
        simplejson.dumps(
            variables_json,
            default=encode_complex,
            ignore_nan=True,
        ),
    ]

    runtime_configuration = variables.get(
        PIPELINE_RUN_MAGE_VARIABLES_KEY,
        {},
    ).get('blocks', {}).get(block.uuid, {}).get('configuration')

    if runtime_configuration:
        if runtime_configuration.get('flags'):
            flags = runtime_configuration['flags']
            flags = flags if type(flags) is list else [flags]
            # e.g. --full-refresh
            args += flags

    if BlockLanguage.SQL == block.language:
        attr = parse_attributes(block)
        project_name = attr['project_name']
        file_path = attr['file_path']
        project_full_path = attr['project_full_path']
        full_path = attr['full_path']
        path_to_model = full_path.replace(f'{project_full_path}{os.sep}', '')

        if test_execution:
            dbt_command = 'compile'

            with open(os.path.join(project_full_path, 'dbt_project.yml')) as f:
                dbt_project = yaml.safe_load(f)
                target_path = dbt_project['target-path']
                path = os.path.join(project_full_path, target_path, 'compiled', file_path)
                if os.path.exists(path):
                    os.remove(path)

        if runtime_configuration:
            prefix = runtime_configuration.get('prefix')
            if prefix:
                path_to_model = f'{prefix}{path_to_model}'

            suffix = runtime_configuration.get('suffix')
            if suffix:
                path_to_model = f'{path_to_model}{suffix}'

        args += [
            '--select',
            path_to_model,
        ]
    else:
        project_name = Template(block.configuration['dbt_project_name']).render(
            variables=variables,
            **get_template_vars(),
        )
        project_full_path = os.path.join(get_repo_path(), 'dbt', project_name)
        args += block.content.split(' ')

    profiles_dir = os.path.join(project_full_path, '.mage_temp_profiles', str(uuid.uuid4()))

    args += [
        '--project-dir',
        project_full_path,
        '--profiles-dir',
        profiles_dir,
    ]

    dbt_profile_target = (block.configuration.get('dbt_profile_target') or
                          variables.get('dbt_profile_target'))

    if dbt_profile_target:
        dbt_profile_target = Template(dbt_profile_target).render(
            variables=lambda x: variables.get(x),
            **get_template_vars(),
        )
        args += [
            '--target',
            dbt_profile_target,
        ]

    return dbt_command, args, dict(
        profile_target=dbt_profile_target,
        profiles_dir=profiles_dir,
        project_full_path=project_full_path,
    )


def create_temporary_profile(project_full_path: str, profiles_dir: str) -> Tuple[str, str]:
    profiles_full_path = os.path.join(project_full_path, PROFILES_FILE_NAME)
    profile = load_profiles_file(profiles_full_path)

    temp_profile_full_path = os.path.join(profiles_dir, PROFILES_FILE_NAME)
    os.makedirs(os.path.dirname(temp_profile_full_path), exist_ok=True)

    with open(temp_profile_full_path, 'w') as f:
        yaml.safe_dump(profile, f)

    return (profile, temp_profile_full_path)


def run_dbt_tests(
    block,
    build_block_output_stdout: Callable[..., object] = None,
    global_vars: Dict = {},
    logger: Logger = None,
    logging_tags: Dict = dict(),
) -> None:
    if logger is not None:
        stdout = StreamToLogger(logger, logging_tags=logging_tags)
    elif build_block_output_stdout:
        stdout = build_block_output_stdout(block.uuid)
    else:
        stdout = sys.stdout

    dbt_command, args, command_line_dict = build_command_line_arguments(
        block,
        global_vars,
        run_tests=True,
    )

    project_full_path = command_line_dict['project_full_path']
    profiles_dir = command_line_dict['profiles_dir']

    _, temp_profile_full_path = create_temporary_profile(
        project_full_path,
        profiles_dir,
    )

    proc1 = subprocess.run([
        'dbt',
        dbt_command,
    ] + args, preexec_fn=os.setsid, stdout=subprocess.PIPE)  # os.setsid doesn't work on Windows

    number_of_errors = 0

    with redirect_stdout(stdout):
        lines = proc1.stdout.decode().split('\n')
        for idx, line in enumerate(lines):
            print(line)

            match = re.search('ERROR=([0-9]+)', line)
            if match:
                number_of_errors += int(match.groups()[0])

    try:
        shutil.rmtree(profiles_dir)
    except Exception as err:
        print(f'Error removing temporary profile at {temp_profile_full_path}: {err}')

    if number_of_errors >= 1:
        raise Exception('DBT test failed.')


def get_dbt_project_settings(block: 'Block') -> Dict:
    attributes_dict = parse_attributes(block)
    dbt_project_full_path = attributes_dict['dbt_project_full_path']

    config = {}
    with open(dbt_project_full_path, 'r') as f:
        config = yaml.safe_load(f)

    return config


def get_model_configurations_from_dbt_project_settings(block: 'Block') -> Dict:
    dbt_project = get_dbt_project_settings(block)

    if not dbt_project.get('models'):
        return

    attributes_dict = parse_attributes(block)
    project_name = attributes_dict['project_name']
    if not dbt_project['models'].get(project_name):
        return

    models_folder_path = attributes_dict['models_folder_path']
    full_path = attributes_dict['full_path']
    parts = full_path.replace(models_folder_path, '').split(os.sep)
    parts = list(filter(lambda x: x, parts))
    if len(parts) >= 2:
        models_subfolder = parts[0]
        if dbt_project['models'][project_name].get(models_subfolder):
            return dbt_project['models'][project_name][models_subfolder]

    return dbt_project['models'][project_name]


def fetch_model_data(
    block: 'Block',
    profile_target: str,
    limit: int = None,
) -> DataFrame:
    attributes_dict = parse_attributes(block)
    model_name = attributes_dict['model_name']
    table_name = attributes_dict['table_name']

    # bigquery: dataset, schema
    # postgres: schema
    # redshift: schema
    # snowflake: schema
    # trino: schema
    profile = get_profile(block, profile_target)
    schema = profile.get('schema') or profile.get('+schema')
    if not schema and 'dataset' in profile:
        schema = profile['dataset']

    if not schema:
        raise print(
            f'WARNING: Cannot fetch data from model {model_name}, ' +
            f'no schema found in profile target {profile_target}.',
        )

    # Check dbt_profiles for schema to append

    # If the model SQL file contains a config with schema, change the schema to use that.
    # https://docs.getdbt.com/reference/resource-configs/schema
    config = model_config(block.content)
    config_schema = config.get('schema')
    if config_schema:
        schema = f'{schema}_{config_schema}'
    else:
        # settings from the dbt_project.yml
        model_configurations = get_model_configurations_from_dbt_project_settings(block)
        model_configuration_schema = None
        if model_configurations:
            model_configuration_schema = (model_configurations.get('schema') or
                                          model_configurations.get('+schema'))

        if model_configuration_schema:
            schema = f"{schema}_{model_configuration_schema}"

    query_string = f'SELECT * FROM {schema}.{table_name}'

    return execute_query(block, profile_target, query_string, limit)


def upstream_blocks_from_sources(block: Block) -> List[Block]:
    mapping = {}
    sources = extract_sources(block.content)
    for tup in sources:
        source_name, table_name = tup
        if source_name not in mapping:
            mapping[source_name] = {}
        mapping[source_name][table_name] = True

    attributes_dict = parse_attributes(block)
    source_name = attributes_dict['source_name']

    arr = []
    for b in block.upstream_blocks:
        table_name = source_table_name_for_block(b)
        if mapping.get(source_name, {}).get(table_name):
            arr.append(b)

    return arr


def model_config(text: str) -> Dict:
    """
    Extract the run time configuration for the model.
    https://docs.getdbt.com/docs/build/custom-aliases
    e.g. {{ config(...) }}
    """
    matches = re.findall(r"""{{\s+config\(([^)]+)\)\s+}}""", text)

    config = {}
    for key_values_string in matches:
        key_values = key_values_string.strip().split(',')
        for key_value_string in key_values:
            parts = key_value_string.strip().split('=')
            if len(parts) == 2:
                key, value = parts
                key = key.strip()
                value = value.strip()
                if value:
                    if ((value[0] == "'" and value[-1] == "'") or
                            (value[0] == '"' and value[-1] == '"')):
                        value = value[1:-1]
                config[key] = value

    return config



from datetime import datetime as dt
from mage_ai.data_cleaner.cleaning_rules.reformat_values import ReformatValues
from mage_ai.data_cleaner.column_types.column_type_detector import infer_column_types
from mage_ai.tests.base_test import TestCase
import numpy as np
import pandas as pd


class ReformatValuesCleaningRuleTests(TestCase):
    def setUp(self):
        self.rng = np.random.default_rng(42)
        return super().setUp()

    def test_standardized_capitalization(self):
        df = pd.DataFrame(
            [
                [None, 'us', 30000, 'Funny Video Corp', 'cute animal #1', 100, 30],
                ['500', 'CA', 10000, 'Machine Learning 4 U', 'intro to regression', 3000, 20],
                ['', '', np.nan, 'News Inc', 'Daily news #1', None, 75],
                ['250', 'CA', 7500, 'Machine Learning 4 U', 'MACHINE LEARNING SEMINAR', 8000, 20],
                ['1000', 'mx', 45003, None, 'cute Animal #4', 90, 40],
                ['1500', 'MX', 75000, 'Funny Video Corp', '', 70, 25],
                ['1500', np.nan, 75000, 'News Inc', 'daily news #3', 70, 25],
                [None, 'mx', 75000, 'Z Combinator', 'Tutorial: how to Start a startup', 70, np.nan],
                ['1250', 'US', 60000, 'Funny Video Corp', 'cute animal #3', 80, 20],
                ['', 'CA', 5000, '', '', 10000, 30],
                ['800', None, 12050, 'Funny Video Corp', 'meme Compilation', 2000, 45],
                ['600', 'CA', 11000, 'News Inc', 'daily news #2', 3000, 50],
                ['600', 'ca', '', 'Funny Video Corp', '', 3000, None],
                ['700', 'MX', 11750, 'Funny Video Corp', 'cute animal #2', 2750, 55],
                ['700', '', None, 'Funny Video Corp', '', None, 55],
                ['700', 'MX', 11750, 'Funny Video Corp', '', 2750, 55],
                ['1200', 'MX', 52000, 'Z Combinator', 'vc funding strats', 75, 60],
            ],
            columns=[
                'userid',
                'location',
                'number_of_creators',
                'company_name',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )
        column_types = {
            'userid': 'category_high_cardinality',
            'location': 'category',
            'company_name': 'category_high_cardinality',
            'number_of_creators': 'number',
            'name': 'text',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        statistics = {
            'userid/count': 13,
            'userid/null_value_rate': 4 / 17,
            'location/count': 13,
            'location/null_value_rate': 4 / 17,
            'company_name/count': 15,
            'company_name/null_value_rate': 2 / 17,
            'number_of_creators/count': 14,
            'number_of_creators/null_value_rate': 3 / 17,
            'name/count': 12,
            'name/null_value_rate': 5 / 17,
            'losses/count': 16,
            'losses/null_value_rate': 1 / 17,
            'number_of_advertisers/count': 16,
            'number_of_advertisers/null_value_rate': 1 / 17,
        }
        rule = ReformatValues(df, column_types, statistics)
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Reformat values',
                message='Format entries in these columns as fully uppercase to improve data '
                        'quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['location'],
                    axis='column',
                    action_options={
                        'reformat': 'caps_standardization',
                        'capitalization': 'uppercase',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Reformat values',
                message='Format entries in these columns as fully lowercase to improve data '
                        'quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['company_name', 'name'],
                    axis='column',
                    action_options={
                        'reformat': 'caps_standardization',
                        'capitalization': 'lowercase',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        self.assertEqual(results, expected_results)

    def test_currency_correction(self):
        df = pd.DataFrame(
            [
                ['$', '$    10000', 'stock exchange america', '$:MAGE', 5.34],
                ['£', '£200', 'huddersfield stock exchange', '£:XYZA', -1.34],
                ['CAD', 'CAD 100', None, '', -0.89],
                ['¥', '¥2500.89', 'stock exchange japan', '', 4.23],
                ['€', '€ 123.34', 'dresden stock exchange', '€:1234', 2.34],
                ['₹', '₹        10000', np.nan, '₹:FDSA', -7.80],
                ['Rs', 'Rs 10000', '', '₹:ASDF', 4.44],
                ['', '10000元', 'stock exchange china', '元:ASDF', 1.02],
                [None, None, 'stock exchange san jose', None, -2.01],
            ],
            columns=[
                'native_currency',
                'value',
                'exchange',
                'ticker',
                'growth_rate',
            ],
        )
        column_types = {
            'native_currency': 'category',
            'value': 'number_with_decimals',
            'exchange': 'category',
            'ticker': 'text',
            'growth_rate': 'number_with_decimals',
        }
        statistics = {
            'native_currency/count': 6,
            'value/count': 8,
            'exchange/count': 6,
            'ticker/count': 6,
            'growth_rate/count': 9,
        }
        rule = ReformatValues(df, column_types, statistics)
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Reformat values',
                message='Format entries in these columns as numbers to improve data quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['value'],
                    axis='column',
                    action_options={
                        'reformat': 'currency_to_num',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        self.assertEqual(results, expected_results)

    def test_currency_conversion_test_all_formatting(self):
        values = [
            '  $ 10000',
            '- ¥ 22.324523',
            'Rs 100000.23   ',
            '  € 12.23425',
            'CAD     12423      ',
            '£ .0927503',
            '-₹ 0',
            ' 10000 元   ',
            ' 0.42 €',
            ' -  3.42032 CAD',
        ]

        df = pd.DataFrame({'column': values})
        column_types = {'column': 'number_with_decimals'}
        statistics = {
            'column/count': 10,
            'column/count_distinct': 10,
            'column/null_value_rate': 0,
        }
        results = ReformatValues(df, column_types, statistics).evaluate()
        expected_results = [
            dict(
                title='Reformat values',
                message='Format entries in these columns as numbers to improve data quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['column'],
                    axis='column',
                    action_options={
                        'reformat': 'currency_to_num',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        self.assertEqual(results, expected_results)

    def test_datetime_conversion(self):
        df = pd.DataFrame(
            [
                [
                    dt(2022, 8, 4),
                    '08/04/22',
                    'Thursday, August 4, 2022',
                    'Thu, Aug 04 22',
                    '8-4-2022',
                ],
                [dt(2022, 1, 20), '', 'Thursday,   JaNUary 20, 2022', 'THU, Jan 20 22', ''],
                [None, '12/24/22', '', 'Sat, Dec 24 2022', '12-24-2022'],
                [
                    dt(2022, 10, 31),
                    '10/31/22',
                    'Monday,   ocTober 31, 2022',
                    'OctobEr is good',
                    None,
                ],
                [dt(2022, 6, 27), None, 'MonDay, June 27, 2022', 'Mon, jUn 27 2022', '6-27-2022'],
                [dt(2022, 3, 8), '03/08/    22', None, 'tuEsday   is a good day', '3-8-2022'],
            ],
            columns=[
                'date1',
                'date2',
                'date3',
                'date4',
                'date5',
            ],
        )
        # date3 and date4 are detected as text type.
        # TODO: Update column_type_detector to detect date3 and date4 as datetime type.
        column_types = infer_column_types(df)
        statistics = {
            'date1/count': 5,
            'date2/count': 4,
            'date3/count': 4,
            'date4/count': 6,
            'date5/count': 4,
        }
        rule = ReformatValues(df, column_types, statistics)
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Reformat values',
                message='Format entries in these columns as datetime objects to improve data '
                        'quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['date2', 'date5'],
                    axis='column',
                    action_options={
                        'reformat': 'date_format_conversion',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        self.assertEqual(results, expected_results)

    def test_datetime_conversion_type_edge_cases(self):
        df = pd.DataFrame(
            [
                [dt(2022, 8, 4), '08/04/22', 'Action Movie #1', 'not a date', 234],
                [dt(2022, 1, 20), '', 'sportsball', '1-20-2022', 13234],
                [None, '12/24/22', 'reality tv show', '12-24-2022', 23234],
                [dt(2022, 10, 31), '10/31/22', '', '10.31.2022', 21432],
                [dt(2022, 6, 27), None, 'action Movie #2', '6/27/2022', 324212],
                [dt(2022, 3, 8), '03/08/    22', 'game show', '3/8/2022', 2034],
            ],
            columns=[
                'date1',
                'date2',
                'notdate',
                'mostlydate',
                'date5',
            ],
        )
        column_types = infer_column_types(df)
        statistics = {
            'date1/count': 5,
            'date2/count': 4,
            'notdate/count': 5,
            'mostlydate/count': 6,
            'date5/count': 6,
            'date1/null_value_rate': 1 / 6,
            'date2/null_value_rate': 2 / 6,
            'notdate/null_value_rate': 1 / 6,
            'mostlydate/null_value_rate': 0 / 6,
            'date5/null_value_rate': 0 / 6,
        }
        rule = ReformatValues(df, column_types, statistics)
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Reformat values',
                message='Format entries in these columns as fully lowercase to improve data '
                        'quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['notdate'],
                    axis='column',
                    action_options={
                        'reformat': 'caps_standardization',
                        'capitalization': 'lowercase',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Reformat values',
                message='Format entries in these columns as datetime objects to improve data '
                        'quality.',
                action_payload=dict(
                    action_type='reformat',
                    action_arguments=['date2', 'mostlydate'],
                    axis='column',
                    action_options={
                        'reformat': 'date_format_conversion',
                    },
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        self.assertEqual(results, expected_results)



from mage_ai.autocomplete.utils import extract_all_imports
from mage_ai.data_preparation.models.constants import BlockType
import re


def convert_to_block(block, content):
    block_template_content = block.content
    content_imports = extract_all_imports(content, ignore_nesting=True)
    block_template_imports = extract_all_imports(block_template_content, ignore_nesting=True)

    for i in content_imports:
        content = content.replace(i, '')
    for i in block_template_imports:
        block_template_content = block_template_content.replace(i, '')

    content = content.strip()
    block_template_content = block_template_content.strip()

    block_template_content_parts = block_template_content.split('\n')
    decorator_index = None
    for i, line in enumerate(block_template_content_parts):
        if decorator_index is not None:
            break
        if re.search(f'@{block.type}', line):
            decorator_index = i

    decorator_function_started = False
    decorator_index_end = None
    block_template_content_parts2 = block_template_content_parts[decorator_index:]
    final_function_in_content = False
    for i, line in enumerate(block_template_content_parts2):
        if decorator_index_end is not None:
            break

        if not decorator_function_started and not re.search(r'^[\w@]+', line):
            decorator_function_started = True

        if decorator_function_started:
            if re.search(r'^[\w]+', line):
                decorator_index_end = i - 1
            elif i == len(block_template_content_parts2) - 1:
                decorator_index_end = i
                final_function_in_content = True

    converted_imports = '\n'.join(sorted(block_template_imports + content_imports)).strip()

    decorator_function_parts = '\n'.join(
        block_template_content_parts[decorator_index:decorator_index + decorator_index_end],
    ).strip().split('\n')

    content_parts = content.split('\n')
    content_to_display = []
    for i, line in enumerate(content_parts):
        if len(content_parts) - 1 == i and \
                block.type in [BlockType.DATA_LOADER, BlockType.TRANSFORMER]:
            content_to_display.append(f'    return {line}')
        else:
            content_to_display.append(f'    {line}')

    if final_function_in_content:
        part_3 = '\n'.join(decorator_function_parts).strip()
        part_5 = ''
    else:
        part_3 = '\n'.join(decorator_function_parts[:-2]).strip()
        part_5 = '\n'.join(
            block_template_content_parts[decorator_index + decorator_index_end:]
        ).strip()

    return """{}


{}


{}

{}


{}""".format(
        converted_imports,
        '\n'.join(block_template_content_parts[:decorator_index]).strip(),
        part_3,
        '\n'.join(content_to_display),
        part_5,
    )



from datetime import datetime
from mage_ai.shared.strings import replacer
from typing import List
import os
import re
import socket


def clean_name(name, allow_characters: List[str] = []):
    for c in ['\ufeff', '\uFEFF', '"', '$', '\n', '\r', '\t']:
        name = name.replace(c, '')

    indexes_of_allowed_characters = {}
    for allowed_char in allow_characters:
        if allowed_char not in indexes_of_allowed_characters:
            indexes_of_allowed_characters[allowed_char] = []

        for idx, char in enumerate(name):
            if char == allowed_char:
                indexes_of_allowed_characters[allowed_char].append(idx)

    name = re.sub(r'\W', '_', name)

    for allowed_char, indexes in indexes_of_allowed_characters.items():
        for idx in indexes:
            name = replacer(name, allowed_char, idx)

    if name and re.match(r'\d', name[0]):
        name = f'letter_{name}'
    return name.lower()


def files_in_path(path, verbose=0):
    files = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(path):
        for file in f:
            files.append(os.path.join(r, file))

    if verbose >= 1:
        for f in files:
            print(f)

    return files


def files_in_single_path(path):
    f = []
    for (dirpath, dirnames, filenames) in os.walk(path):
        f.extend([os.path.join(dirpath, file) for file in filenames])
        break
    return f


def convert_pandas_dtype_to_python_type(dtype):
    dtype = str(dtype)
    if dtype in ['int64', 'integer']:
        return int
    elif 'float' in dtype:
        return float
    elif dtype == 'bool':
        return bool
    elif 'datetime' in dtype:
        return datetime
    return str


def convert_python_type_to_redshift_type(python_type):
    if python_type is int:
        return 'BIGINT'
    elif python_type is float:
        return 'FLOAT8'
    elif python_type is bool:
        return 'BOOLEAN'
    return 'VARCHAR'


def convert_python_type_to_bigquery_type(python_type):
    if python_type is int:
        return 'INT64'
    elif python_type is float:
        return 'FLOAT64'
    elif python_type is bool:
        return 'BOOL'
    elif python_type is datetime:
        return 'DATETIME'
    return 'STRING'


def convert_python_type_to_trino_type(python_type):
    if python_type is int:
        return 'BIGINT'
    elif python_type is float:
        return 'DOUBLE'
    elif python_type is bool:
        return 'BOOLEAN'
    elif python_type is datetime:
        return 'TIMESTAMP'
    return 'VARCHAR'


def is_port_in_use(port: int) -> bool:
    print(f'Checking port {port}...')
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def is_spark_env():
    import importlib
    return importlib.util.find_spec('pyspark') is not None



from mage_ai.streaming.sinks.opensearch import OpenSearchSink
from mage_ai.streaming.sinks.sink_factory import SinkFactory
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class SinkFactoryTests(TestCase):
    def test_get_source_kafka(self):
        with patch.object(OpenSearchSink, '__init__', return_value=None) as mock_init:
            config = dict(
                connector_type='opensearch',
            )
            sink = SinkFactory.get_sink(config)
            self.assertIsInstance(sink, OpenSearchSink)
            mock_init.assert_called_once_with(config)

    def test_get_source_other(self):
        with self.assertRaises(Exception) as context:
            SinkFactory.get_sink(dict(
                connector_type='random',
            ))
        self.assertTrue(
            'Ingesting data to random is not supported in streaming pipelines yet.'
            in str(context.exception),
        )



from mage_ai.services.aws.s3 import s3
from mage_ai.data_preparation.storage.base_storage import BaseStorage
from mage_ai.shared.constants import S3_PREFIX
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.urls import s3_url_path
from typing import Dict, List
import io
import json
import pandas as pd
import polars as pl
import simplejson


class S3Storage(BaseStorage):
    def __init__(self, bucket=None, dirpath=None, **kwargs):
        if bucket is not None:
            self.client = s3.Client(bucket, **kwargs)
        else:
            if dirpath is None or not dirpath.startswith(S3_PREFIX):
                raise Exception('')
            path_parts = dirpath.replace(S3_PREFIX, '').split('/')
            bucket = path_parts.pop(0)
            self.client = s3.Client(bucket, **kwargs)

    def isdir(self, path: str) -> bool:
        if not path.endswith('/'):
            path += '/'
        return self.path_exists(path)

    def listdir(self, path: str, suffix: str = None) -> List[str]:
        if not path.endswith('/'):
            path += '/'
        path = s3_url_path(path)
        keys = self.client.listdir(path, suffix=suffix)
        return [k[len(path):].rstrip('/') for k in keys]

    def makedirs(self, path: str, **kwargs) -> None:
        pass

    def path_exists(self, path: str) -> bool:
        results = self.client.list_objects(s3_url_path(path), max_keys=1)
        return len(results) > 0

    def remove(self, path: str) -> None:
        self.client.delete_objects(s3_url_path(path))

    def remove_dir(self, path: str) -> None:
        self.client.delete_objects(s3_url_path(path))

    def read_json_file(self, file_path: str, default_value={}) -> Dict:
        try:
            return json.loads(self.client.read(s3_url_path(file_path)))
        except Exception:
            return default_value

    async def read_json_file_async(self, file_path: str, default_value={}) -> Dict:
        """
        TODO: Implement async http call.
        """
        return self.read_json_file(file_path, default_value=default_value)

    def write_json_file(self, file_path: str, data) -> None:
        self.client.upload(
            s3_url_path(file_path),
            simplejson.dumps(
                data,
                default=encode_complex,
                ignore_nan=True,
            ),
        )

    async def write_json_file_async(self, file_path: str, data) -> None:
        """
        TODO: Implement async http call.
        """
        return self.write_json_file(file_path, data)

    def read_parquet(self, file_path: str, **kwargs) -> pd.DataFrame:
        buffer = io.BytesIO(self.client.get_object(s3_url_path(file_path)).read())
        return pd.read_parquet(buffer, **kwargs)

    def write_parquet(self, df: pd.DataFrame, file_path: str) -> None:
        buffer = io.BytesIO()
        df.to_parquet(buffer)
        buffer.seek(0)
        self.client.upload_object(s3_url_path(file_path), buffer)

    def write_polars_dataframe(self, df: pl.DataFrame, file_path: str) -> None:
        buffer = io.BytesIO()
        df.write_parquet(buffer)
        buffer.seek(0)
        self.client.upload_object(s3_url_path(file_path), buffer)



from botocore.config import Config
from mage_ai.services.aws.ecs.config import EcsConfig
import boto3
import json
import os


def run_task(
    command: str,
    ecs_config: EcsConfig,
    wait_for_completion: bool = True,
) -> None:
    if type(ecs_config) is dict:
        ecs_config = EcsConfig.load(config=ecs_config)
    client = boto3.client('ecs')
    response = client.run_task(**ecs_config.get_task_config(command=command))

    if wait_for_completion:
        arn = response['tasks'][0]['taskArn']
        waiter = client.get_waiter('tasks_stopped')
        waiter.wait(cluster=ecs_config.cluster, tasks=[arn])

    print(json.dumps(response, indent=4, default=str))
    return response


def stop_task(task_arn: str, cluster: str = None) -> None:
    client = boto3.client('ecs')
    return client.stop_task(
        cluster=cluster,
        task=task_arn,
    )


def list_tasks(cluster):
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    ecs_client = boto3.client('ecs', config=config)

    task_arns = ecs_client.list_tasks(
        cluster=cluster,
    )['taskArns']

    return ecs_client.describe_tasks(
        cluster=cluster,
        include=[
            'TAGS',
        ],
        tasks=task_arns,
    )



from mage_ai.data_cleaner.cleaning_rules.remove_collinear_columns import RemoveCollinearColumns
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.tests.base_test import TestCase
from pandas.testing import assert_frame_equal
import numpy as np
import pandas as pd


class RemoveCollinearColumnsTests(TestCase):
    def setUp(self):
        self.rng = np.random.default_rng(42)
        return super().setUp()

    def test_categorical_data_frame(self):
        df = pd.DataFrame(
            [
                [1, 1000, '2021-10-01', '2021-09-01'],
                [1, 1050, '2021-10-01', '2021-08-01'],
                [1, 1100, '2021-10-01', '2021-01-01'],
                [2, 1150, '2021-09-01', '2021-08-01'],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
            ],
        )
        column_types = {
            'group_id': 'category',
            'order_id': 'category',
            'group_churned_at': 'datetime',
            'order_created_at': 'datetime',
        }
        statistics = {}
        df = clean_dataframe(df, column_types, dropna=False)
        result = RemoveCollinearColumns(df, column_types, statistics).evaluate()
        self.assertEqual(result, [])

    def test_clean_removes_all_data_frame(self):
        df = pd.DataFrame(
            [
                [None, 1000, '2021-10-01', '2021-09-01'],
                [1, None, '2021-10-01', '2021-08-01'],
                [np.nan, 1100, '2021-10-01', '2021-01-01'],
                [None, 1150, '2021-09-01', '2021-08-01'],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
            ],
        )
        column_types = {
            'group_id': 'number',
            'order_id': 'number',
            'group_churned_at': 'datetime',
            'order_created_at': 'datetime',
        }
        statistics = {}
        df = clean_dataframe(df, column_types, dropna=False)
        result = RemoveCollinearColumns(df, column_types, statistics).evaluate()
        self.assertEqual(result, [])

    def test_evaluate(self):
        df = pd.DataFrame(
            [
                [1000, 30000, 10, 100, 30],
                [500, 10000, 20, 3000, 20],
                [250, 7500, 25, 8000, 20],
                [1000, 45003, 20, 90, 40],
                [1500, 75000, 30, 70, 25],
                [1250, 60000, 50, 80, 20],
                [200, 5000, 30, 10000, 30],
                [800, 12050, 40, 2000, 45],
                [600, 11000, 50, 3000, 50],
                [700, 11750, 20, 2750, 55],
                [1200, 52000, 10, 75, 60],
            ],
            columns=[
                'number_of_users',
                'views',
                'number_of_creators',
                'losses',
                'number_of_advertisers',
            ],
        )
        column_types = {
            'number_of_users': 'number',
            'views': 'number',
            'number_of_creators': 'number',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        statistics = {}
        df = clean_dataframe(df, column_types, dropna=False)
        rule = RemoveCollinearColumns(df, column_types, statistics)
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Remove collinear columns',
                message='Delete these columns to remove redundant data and increase data quality.',
                status='not_applied',
                action_payload=dict(
                    action_type='remove',
                    action_arguments=['number_of_users'],
                    axis='column',
                    action_options={},
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
            )
        ]
        self.assertEqual(results, expected_results)

    def test_evaluate_bad_dtypes(self):
        df = pd.DataFrame(
            [
                [1000, 'US', 30000, '10', 'cute animal #1', 100, '30'],
                ['500', 'CA', 10000, '20', 'intro to regression', 3000, '20'],
                [200, '', np.nan, '50', 'daily news #1', None, '75'],
                [250, 'CA', 7500, '25', 'machine learning seminar', 8000, '20'],
                ['1000', 'MX', 45003, '20', 'cute animal #4', 90, '40'],
                [1500, 'MX', 75000, '30', '', 70, '25'],
                [1500, 'US', 75000, np.nan, 'daily news #3', 70, '25'],
                [None, 'US', 75000, '30', 'tutorial: how to start a startup', 70, np.nan],
                [1250, 'US', 60000, '50', 'cute animal #3', 80, '20'],
                ['200', 'CA', 5000, '30', '', 10000, '30'],
                [800, 'US', 12050, '40', 'meme compilation', 2000, '45'],
                ['600', 'CA', 11000, '50', 'daily news #2', 3000, '50'],
                [600, 'CA', '', 50, '', 3000, None],
                ['700', 'MX', 11750, '20', 'cute animal #2', 2750, '55'],
                [700, '', None, 20, '', None, '55'],
                [700, 'MX', 11750, '', '', 2750, '55'],
                [1200, 'MX', 52000, '10', 'vc funding strats', 75, '60'],
            ],
            columns=[
                'number_of_users',
                'location',
                'views',
                'number_of_creators',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )
        cleaned_df = pd.DataFrame(
            [
                [1000, 30000, 10, 100, 30],
                [500, 10000, 20, 3000, 20],
                [250, 7500, 25, 8000, 20],
                [1000, 45003, 20, 90, 40],
                [1500, 75000, 30, 70, 25],
                [1250, 60000, 50, 80, 20],
                [200, 5000, 30, 10000, 30],
                [800, 12050, 40, 2000, 45],
                [600, 11000, 50, 3000, 50],
                [700, 11750, 20, 2750, 55],
                [1200, 52000, 10, 75, 60],
            ],
            columns=[
                'number_of_users',
                'views',
                'number_of_creators',
                'losses',
                'number_of_advertisers',
            ],
        ).astype(float)
        column_types = {
            'number_of_users': 'number',
            'location': 'category',
            'views': 'number',
            'number_of_creators': 'number',
            'name': 'text',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        statistics = {}
        df = clean_dataframe(df, column_types)
        rule = RemoveCollinearColumns(df, column_types, statistics)
        assert_frame_equal(cleaned_df, rule.numeric_df.reset_index(drop=True))
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Remove collinear columns',
                message='Delete these columns to remove redundant data and increase data quality.',
                status='not_applied',
                action_payload=dict(
                    action_type='remove',
                    action_arguments=['number_of_users'],
                    axis='column',
                    action_options={},
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
            )
        ]
        self.assertEqual(results, expected_results)

    def test_evaluate_dirty(self):
        df = pd.DataFrame(
            [
                [1000, 30000, 10, 100, 30],
                [500, 10000, 20, 3000, 20],
                [200, np.nan, 50, None, 75],
                [250, 7500, 25, 8000, 20],
                [1000, 45003, 20, 90, 40],
                [1500, 75000, 30, 70, 25],
                [1500, 75000, np.nan, 70, 25],
                [None, 75000, 30, 70, np.nan],
                [1250, 60000, 50, 80, 20],
                [200, 5000, 30, 10000, 30],
                [800, 12050, 40, 2000, 45],
                [600, 11000, 50, 3000, 50],
                [600, '', 50, 3000, None],
                [700, 11750, 20, 2750, 55],
                [700, None, 20, None, 55],
                [700, 11750, '', 2750, 55],
                [1200, 52000, 10, 75, 60],
            ],
            columns=[
                'number_of_users',
                'views',
                'number_of_creators',
                'losses',
                'number_of_advertisers',
            ],
        )
        cleaned_df = pd.DataFrame(
            [
                [1000, 30000, 10, 100, 30],
                [500, 10000, 20, 3000, 20],
                [250, 7500, 25, 8000, 20],
                [1000, 45003, 20, 90, 40],
                [1500, 75000, 30, 70, 25],
                [1250, 60000, 50, 80, 20],
                [200, 5000, 30, 10000, 30],
                [800, 12050, 40, 2000, 45],
                [600, 11000, 50, 3000, 50],
                [700, 11750, 20, 2750, 55],
                [1200, 52000, 10, 75, 60],
            ],
            columns=[
                'number_of_users',
                'views',
                'number_of_creators',
                'losses',
                'number_of_advertisers',
            ],
        ).astype(float)
        column_types = {
            'number_of_users': 'number',
            'views': 'number',
            'number_of_creators': 'number',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        statistics = {}
        df = clean_dataframe(df, column_types, dropna=False)
        rule = RemoveCollinearColumns(df, column_types, statistics)
        assert_frame_equal(cleaned_df, rule.numeric_df.reset_index(drop=True))
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Remove collinear columns',
                message='Delete these columns to remove redundant data and increase data quality.',
                status='not_applied',
                action_payload=dict(
                    action_type='remove',
                    action_arguments=['number_of_users'],
                    axis='column',
                    action_options={},
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
            )
        ]
        self.assertEqual(results, expected_results)

    def test_evaluate_non_numeric(self):
        df = pd.DataFrame(
            [
                [1000, 'US', 30000, 10, 'cute animal #1', 100, 30],
                [500, 'CA', 10000, 20, 'intro to regression', 3000, 20],
                [200, '', np.nan, 50, 'daily news #1', None, 75],
                [250, 'CA', 7500, 25, 'machine learning seminar', 8000, 20],
                [1000, 'MX', 45003, 20, 'cute animal #4', 90, 40],
                [1500, 'MX', 75000, 30, '', 70, 25],
                [1500, 'US', 75000, np.nan, 'daily news #3', 70, 25],
                [None, 'US', 75000, 30, 'tutorial: how to start a startup', 70, np.nan],
                [1250, 'US', 60000, 50, 'cute animal #3', 80, 20],
                [200, 'CA', 5000, 30, '', 10000, 30],
                [800, 'US', 12050, 40, 'meme compilation', 2000, 45],
                [600, 'CA', 11000, 50, 'daily news #2', 3000, 50],
                [600, 'CA', '', 50, '', 3000, None],
                [700, 'MX', 11750, 20, 'cute animal #2', 2750, 55],
                [700, '', None, 20, '', None, 55],
                [700, 'MX', 11750, '', '', 2750, 55],
                [1200, 'MX', 52000, 10, 'vc funding strats', 75, 60],
            ],
            columns=[
                'number_of_users',
                'location',
                'views',
                'number_of_creators',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )
        cleaned_df = pd.DataFrame(
            [
                [1000, 30000, 10, 100, 30],
                [500, 10000, 20, 3000, 20],
                [250, 7500, 25, 8000, 20],
                [1000, 45003, 20, 90, 40],
                [1500, 75000, 30, 70, 25],
                [1250, 60000, 50, 80, 20],
                [200, 5000, 30, 10000, 30],
                [800, 12050, 40, 2000, 45],
                [600, 11000, 50, 3000, 50],
                [700, 11750, 20, 2750, 55],
                [1200, 52000, 10, 75, 60],
            ],
            columns=[
                'number_of_users',
                'views',
                'number_of_creators',
                'losses',
                'number_of_advertisers',
            ],
        ).astype(float)
        column_types = {
            'number_of_users': 'number',
            'location': 'category',
            'views': 'number',
            'number_of_creators': 'number',
            'name': 'text',
            'losses': 'number',
            'number_of_advertisers': 'number',
        }
        statistics = {}
        df = clean_dataframe(df, column_types, dropna=False)
        rule = RemoveCollinearColumns(df, column_types, statistics)
        assert_frame_equal(cleaned_df, rule.numeric_df.reset_index(drop=True))
        results = rule.evaluate()
        expected_results = [
            dict(
                title='Remove collinear columns',
                message='Delete these columns to remove redundant data and increase data quality.',
                status='not_applied',
                action_payload=dict(
                    action_type='remove',
                    action_arguments=['number_of_users'],
                    axis='column',
                    action_options={},
                    action_variables={},
                    action_code='',
                    outputs=[],
                ),
            )
        ]
        self.assertEqual(results, expected_results)

    # TODO: Make this test case deterministic
    # def test_perfectly_collinear(self):
    #     number_of_users = self.rng.integers(1000, 500000, (10000))
    #     views = number_of_users * 300
    #     revenue = 2 * views - number_of_users
    #     losses = revenue / views + number_of_users
    #     df = pd.DataFrame(
    #         {
    #             'number_of_users': number_of_users,
    #             'views': views,
    #             'revenue': revenue,
    #             'losses': losses,
    #         }
    #     )
    #     column_types = {
    #         'number_of_users': 'number',
    #         'views': 'number',
    #         'revenue': 'number',
    #         'losses': 'number',
    #     }
    #     statistics = {}
    #     df = clean_dataframe(df, column_types, dropna=False)
    #     result = RemoveCollinearColumns(df, column_types, statistics).evaluate()
    #     expected_results = [
    #         dict(
    #             title='Remove collinear columns',
    #             message='Delete these columns to remove redundant data and increase '
    #                     'data quality.',
    #             status='not_applied',
    #             action_payload=dict(
    #                 action_type='remove',
    #                 action_arguments=['revenue', 'views'],
    #                 axis='column',
    #                 action_options={},
    #                 action_variables={},
    #                 action_code='',
    #                 outputs=[],
    #             ),
    #         )
    #     ]
    #     self.assertEqual(result, expected_results)



"""Add sla to pipeline schedule

Revision ID: 84de4cdd6126
Revises: b01be687e537
Create Date: 2022-10-21 11:13:27.570829

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '84de4cdd6126'
down_revision = 'b01be687e537'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_schedule', sa.Column('sla', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_schedule', 'sla')
    # ### end Alembic commands ###



from dataclasses import dataclass
from enum import Enum
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.constants import DEFAULT_BATCH_SIZE
from mage_ai.streaming.sources.base import BaseSource
from mage_ai.streaming.sources.shared import SerializationMethod, SerDeConfig
from typing import Callable, Dict
import boto3
import json

DEFAULT_WAIT_TIME_SECONDS = 1


class MessageDeletionMethod(str, Enum):
    AFTER_RECEIVED = 'AFTER_RECEIVED'
    MANUAL = 'MANUAL'


@dataclass
class AmazonSqsConfig(BaseConfig):
    queue_name: str
    batch_size: int = DEFAULT_BATCH_SIZE
    wait_time_seconds: int = DEFAULT_WAIT_TIME_SECONDS
    serde_config: SerDeConfig = None
    message_deletion_method: MessageDeletionMethod = MessageDeletionMethod.AFTER_RECEIVED

    @classmethod
    def parse_config(self, config: Dict) -> Dict:
        serde_config = config.get('serde_config')
        if serde_config and type(serde_config) is dict:
            config['serde_config'] = SerDeConfig(**serde_config)
        return config


class AmazonSqsSource(BaseSource):
    config_class = AmazonSqsConfig

    def init_client(self):
        self._print('Start initializing consumer.')
        # Initialize kafka consumer
        self.sqs_client = boto3.resource('sqs')
        self.queue = self.sqs_client.get_queue_by_name(QueueName=self.config.queue_name)
        # self.details = self.kinesis_client.describe_stream(
        #     StreamName=self.config.stream_name,
        # )['StreamDescription']
        print(self.queue.url)
        self._print('Finish initializing consumer.')

    def read(self, handler: Callable):
        pass

    def batch_read(self, handler: Callable):
        self._print('Start consuming messages.')
        if self.config.batch_size > 0:
            batch_size = self.config.batch_size
        else:
            batch_size = DEFAULT_BATCH_SIZE

        try:
            while True:
                messages = self.queue.receive_messages(
                    MessageAttributeNames=['All'],
                    MaxNumberOfMessages=batch_size,
                    WaitTimeSeconds=self.config.wait_time_seconds,
                )
                parsed_messages = []
                for msg in messages:
                    parsed_msg_body = self.__deserialize_message(msg.body)
                    if self.config.message_deletion_method == MessageDeletionMethod.AFTER_RECEIVED:
                        parsed_messages.append(parsed_msg_body)
                        msg.delete()
                    else:
                        parsed_messages.append(dict(
                            parsed_msg_body=self.__deserialize_message(msg.body),
                            raw_message=msg,
                        ))
                if len(parsed_messages) > 0:
                    self._print(f'Received {len(parsed_messages)} message. '
                                f'Sample: {parsed_messages[0]}.')
                    handler(parsed_messages)
        except Exception:
            self._print(f'Couldn\'t receive messages from queue {self.config.queue_name}.')
            raise

    def __deserialize_message(self, message):
        if self.config.serde_config is not None and \
                self.config.serde_config.serialization_method == SerializationMethod.RAW_VALUE:
            return message
        else:
            return json.loads(message)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.variable import VariableType
from mage_ai.orchestration.db import safe_db_query


class BlockOutputResource(GenericResource):
    """
    Resource to fetch block output for the notebook. Created to support legacy
    endpoint /api/pipelines/<pipeline_uuid>/blocks/<block_uuid>/outputs
    """
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        block_uuid = pk

        query = kwargs.get('query', {})
        pipeline_uuid = query.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]

        outputs = []
        if pipeline_uuid is not None:
            pipeline = Pipeline.get(pipeline_uuid)
            block = pipeline.get_block(block_uuid)
            error = ApiError.RESOURCE_ERROR.copy()
            if block is None:
                error.update(
                    message=f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
                raise ApiError(error)
            # Only fetch dataframe variables by default
            outputs = block.get_outputs(
                include_print_outputs=False,
                sample_count=None,
                variable_type=VariableType.DATAFRAME,
            )

        return self(dict(outputs=outputs), user, **kwargs)



{% extends "data_loaders/default.jinja" %}
{% block imports %}
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.bigquery import BigQuery
from mage_ai.io.config import ConfigFileLoader
from os import path
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_data_from_big_query(*args, **kwargs):
    """
    Template for loading data from a BigQuery warehouse.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#bigquery
    """
    query = 'your_gbq_query'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    return BigQuery.with_config(ConfigFileLoader(config_path, config_profile)).load(query)
{% endblock %}



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.VariablePresenter import VariablePresenter


class VariablePolicy(BasePolicy):
    pass


VariablePolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

VariablePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

VariablePolicy.allow_read(VariablePresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

VariablePolicy.allow_read(VariablePresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())

VariablePolicy.allow_write([
    'name',
    'value',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_notebook_edit_access())



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.preferences import (
    get_preferences,
    Preferences,
)
from mage_ai.data_preparation.shared.secrets import create_secret
from mage_ai.data_preparation.sync import (
    GitConfig,
    GIT_ACCESS_TOKEN_SECRET_NAME,
    GIT_SSH_PRIVATE_KEY_SECRET_NAME,
    GIT_SSH_PUBLIC_KEY_SECRET_NAME,
)
from mage_ai.data_preparation.sync.git_sync import GitSync
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.orchestration.db.models.secrets import Secret
import os


def get_ssh_public_key_secret_name(user: User = None) -> str:
    return f'{GIT_SSH_PUBLIC_KEY_SECRET_NAME}_{user.id}' if user \
        else GIT_SSH_PUBLIC_KEY_SECRET_NAME


def get_ssh_private_key_secret_name(user: User = None) -> str:
    return f'{GIT_SSH_PRIVATE_KEY_SECRET_NAME}_{user.id}' if user \
        else GIT_SSH_PRIVATE_KEY_SECRET_NAME


def get_access_token_secret_name(user: User = None) -> str:
    return f'{GIT_ACCESS_TOKEN_SECRET_NAME}_{user.id}' if user \
        else GIT_ACCESS_TOKEN_SECRET_NAME


class SyncResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        preferences = get_preferences(user=user)
        sync_config = preferences.sync_config
        return self.build_result_set(
            [sync_config],
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        ssh_public_key = payload.pop('ssh_public_key', None)
        ssh_private_key = payload.pop('ssh_private_key', None)

        if ssh_public_key:
            secret_name = get_ssh_public_key_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, ssh_public_key)
            payload['ssh_public_key_secret_name'] = secret_name
        if ssh_private_key:
            secret_name = get_ssh_private_key_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, ssh_private_key)
            payload['ssh_private_key_secret_name'] = secret_name

        access_token = payload.pop('access_token', None)
        if access_token:
            secret_name = get_access_token_secret_name(user=user)
            secret = Secret.query.filter(
                Secret.name == secret_name).one_or_none()
            if secret:
                secret.delete()
            create_secret(secret_name, access_token)
            payload['access_token_secret_name'] = secret_name

        preferences = Preferences(user=user) if user else get_preferences()
        updated_config = dict(preferences.sync_config, **payload)
        # default repo_path to os.getcwd()
        if not updated_config.get('repo_path', None):
            updated_config['repo_path'] = os.getcwd()
        # Validate payload
        sync_config = GitConfig.load(config=updated_config)

        preferences.update_preferences(dict(sync_config=updated_config))

        GitSync(sync_config)

        return self(get_preferences(user=user).sync_config, user, **kwargs)

    @classmethod
    def member(self, pk, user, **kwargs):
        return self(get_preferences(user=user).sync_config, user, **kwargs)

    def update(self, payload, **kwargs):
        config = GitConfig.load(config=self.model)
        sync = GitSync(config)
        action_type = payload.get('action_type')
        if action_type == 'sync_data':
            sync.sync_data()

        return self



"""add event_variables and metadata to pipeline_runs

Revision ID: 8971d4cd5b39
Revises: 2266370f589b
Create Date: 2022-12-01 00:07:24.890239

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8971d4cd5b39'
down_revision = '2266370f589b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pipeline_run', sa.Column('event_variables', sa.JSON(), nullable=True))
    op.add_column('pipeline_run', sa.Column('metrics', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('pipeline_run', 'event_variables')
    op.drop_column('pipeline_run', 'metrics')
    # ### end Alembic commands ###



from alembic import context
from logging.config import fileConfig
from sqlalchemy import engine_from_config
from sqlalchemy import pool
import logging
import os
import sys

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

logging.getLogger('alembic').setLevel(config.get_section_option('logger_alembic', 'level'))

# add your model's MetaData object here for 'autogenerate' support


sys.path.insert(0, os.path.dirname(
    # mage_ai
    os.path.dirname(
        # orchestration
        os.path.dirname(
            # db
            os.path.dirname(
                # migrations
                os.path.dirname(os.path.abspath(__file__)))))))

from mage_ai.orchestration.db.models.base import Base  # noqa: E402
import mage_ai.orchestration.db.models.oauth  # noqa: E402, F401
import mage_ai.orchestration.db.models.schedules  # noqa: E402, F401
import mage_ai.orchestration.db.models.secrets  # noqa: E402, F401

target_metadata = Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        # https://stackoverflow.com/questions/30378233/sqlite-lack-of-alter-support-alembic-migration-failing-because-of-this-solutio
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            # https://stackoverflow.com/questions/30378233/sqlite-lack-of-alter-support-alembic-migration-failing-because-of-this-solutio
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from enum import Enum
from jupyter_client import KernelManager
from jupyter_client.session import Session

from mage_ai.data_preparation.models.constants import PipelineType


class KernelName(str, Enum):
    PYSPARK = 'pysparkkernel'
    PYTHON3 = 'python3'


PIPELINE_TO_KERNEL_NAME = {
    PipelineType.INTEGRATION: KernelName.PYTHON3,
    PipelineType.DATABRICKS: KernelName.PYTHON3,
    PipelineType.PYTHON: KernelName.PYTHON3,
    PipelineType.PYSPARK: KernelName.PYSPARK,
    PipelineType.STREAMING: KernelName.PYTHON3,
}


DEFAULT_KERNEL_NAME = KernelName.PYTHON3


kernel_managers = dict(
    python3=KernelManager(
        session=Session(key=bytes()),
    ),
    pysparkkernel=KernelManager(
        kernel_name='pysparkkernel',
        session=Session(key=bytes()),
    ),
)



from enum import Enum

PIPELINES_FOLDER = 'pipelines'
PIPELINE_CONFIG_FILE = 'metadata.yaml'
PREFERENCES_FILE = '.preferences.yaml'
DATA_INTEGRATION_CATALOG_FILE = 'data_integration_catalog.json'


DATAFRAME_ANALYSIS_KEYS = frozenset(
    [
        'metadata',
        'statistics',
        'insights',
        'suggestions',
    ]
)
DATAFRAME_ANALYSIS_MAX_ROWS = 100_000
DATAFRAME_ANALYSIS_MAX_COLUMNS = 100
DATAFRAME_SAMPLE_COUNT_PREVIEW = 10
DATAFRAME_SAMPLE_COUNT = 1000
DATAFRAME_SAMPLE_MAX_COLUMNS = 1000
VARIABLE_DIR = '.variables'
LOGS_DIR = '.logs'


class BlockLanguage(str, Enum):
    MARKDOWN = 'markdown'
    PYTHON = 'python'
    R = 'r'
    SQL = 'sql'
    YAML = 'yaml'


class BlockStatus(str, Enum):
    EXECUTED = 'executed'
    FAILED = 'failed'
    NOT_EXECUTED = 'not_executed'
    UPDATED = 'updated'


class BlockType(str, Enum):
    CALLBACK = 'callback'
    CHART = 'chart'
    CUSTOM = 'custom'
    DATA_EXPORTER = 'data_exporter'
    DATA_LOADER = 'data_loader'
    DBT = 'dbt'
    EXTENSION = 'extension'
    MARKDOWN = 'markdown'
    SCRATCHPAD = 'scratchpad'
    SENSOR = 'sensor'
    TRANSFORMER = 'transformer'


class BlockColor(str, Enum):
    BLUE = 'blue'
    GREY = 'grey'
    PINK = 'pink'
    PURPLE = 'purple'
    TEAL = 'teal'
    YELLOW = 'yellow'


class CallbackStatus(str, Enum):
    FAILURE = 'failure'
    SUCCESS = 'success'


class ExecutorType(str, Enum):
    LOCAL_PYTHON = 'local_python'
    ECS = 'ecs'
    GCP_CLOUD_RUN = 'gcp_cloud_run'
    K8S = 'k8s'
    PYSPARK = 'pyspark'

    @classmethod
    def is_valid_type(cls, executor_type: str) -> bool:
        return executor_type.upper() in cls.__members__


class PipelineType(str, Enum):
    INTEGRATION = 'integration'
    DATABRICKS = 'databricks'
    PYTHON = 'python'
    PYSPARK = 'pyspark'
    STREAMING = 'streaming'


class PipelineStatus(str, Enum):
    ACTIVE = 'active',              # At least one active trigger
    INACTIVE = 'inactive',          # All inactive triggers
    NO_SCHEDULES = 'no_schedules',  # No triggers


BLOCK_LANGUAGE_TO_FILE_EXTENSION = {
    BlockLanguage.MARKDOWN: 'md',
    BlockLanguage.PYTHON: 'py',
    BlockLanguage.R: 'r',
    BlockLanguage.SQL: 'sql',
    BlockLanguage.YAML: 'yaml',
}


CALLBACK_STATUSES = [
    CallbackStatus.FAILURE,
    CallbackStatus.SUCCESS,
]


FILE_EXTENSION_TO_BLOCK_LANGUAGE = {
    'md': BlockLanguage.MARKDOWN,
    'py': BlockLanguage.PYTHON,
    'r': BlockLanguage.R,
    'sql': BlockLanguage.SQL,
    'yaml': BlockLanguage.YAML,
}


CUSTOM_EXECUTION_BLOCK_TYPES = [
    BlockType.CHART,
    BlockType.CUSTOM,
    BlockType.DATA_EXPORTER,
    BlockType.DATA_LOADER,
    BlockType.DBT,
    BlockType.EXTENSION,
    BlockType.SENSOR,
    BlockType.TRANSFORMER,
]

NON_PIPELINE_EXECUTABLE_BLOCK_TYPES = [
    BlockType.CHART,
    BlockType.MARKDOWN,
    BlockType.SCRATCHPAD,
]



from mage_ai.data_preparation.repo_manager import set_repo_path
from mage_ai.orchestration.db import TEST_DB, db_connection
from mage_ai.orchestration.db.database_manager import database_manager
from mage_ai.shared.logger import LoggingLevel
import os
import shutil
import sys
import unittest


if sys.version_info.major <= 3 and sys.version_info.minor <= 7:
    class AsyncDBTestCase():
        pass
else:
    class AsyncDBTestCase(unittest.IsolatedAsyncioTestCase):
        def setUp(self):
            pass

        def tearDown(self):
            pass

        @classmethod
        def setUpClass(self):
            super().setUpClass()
            self.repo_path = os.getcwd() + '/test'
            set_repo_path(self.repo_path)
            if not os.path.exists(self.repo_path):
                os.mkdir(self.repo_path)
            database_manager.run_migrations(log_level=LoggingLevel.ERROR)
            db_connection.start_session()

        @classmethod
        def tearDownClass(self):
            shutil.rmtree(self.repo_path)
            db_connection.close_session()

            if os.path.isfile(TEST_DB):
                os.remove(TEST_DB)

            super().tearDownClass()


class DBTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.repo_path = os.getcwd() + '/test'
        set_repo_path(self.repo_path)
        if not os.path.exists(self.repo_path):
            os.mkdir(self.repo_path)
        database_manager.run_migrations(log_level=LoggingLevel.ERROR)
        db_connection.start_session()

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.repo_path)
        db_connection.close_session()

        if os.path.isfile(TEST_DB):
            os.remove(TEST_DB)

        super().tearDownClass()


class TestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass



# ECS environment variables
ECS_CLUSTER_NAME = 'ECS_CLUSTER_NAME'
ECS_TASK_DEFINITION = 'ECS_TASK_DEFINITION'
ECS_CONTAINER_NAME = 'ECS_CONTAINER_NAME'

# GCP environment variables
GCP_PROJECT_ID = 'GCP_PROJECT_ID'
GCP_PATH_TO_KEYFILE = 'path_to_keyfile'
GCP_REGION = 'GCP_REGION'

# K8s environment variables
KUBE_NAMESPACE = 'KUBE_NAMESPACE'
KUBE_SERVICE_TYPE = 'KUBE_SERVICE_TYPE'
KUBE_SERVICE_GCP_BACKEND_CONFIG = 'KUBE_SERVICE_GCP_BACKEND_CONFIG'
KUBE_STORAGE_CLASS_NAME = 'KUBE_STORAGE_CLASS_NAME'
KUBE_SERVICE_ACCOUNT_NAME = 'KUBE_SERVICE_ACCOUNT_NAME'
CLOUD_SQL_CONNECTION_NAME = 'CLOUD_SQL_CONNECTION_NAME'
CONNECTION_URL_SECRETS_NAME = 'CONNECTION_URL_SECRETS_NAME'
DB_SECRETS_NAME = 'DB_SECRETS_NAME'
SERVICE_ACCOUNT_SECRETS_NAME = 'SERVICE_ACCOUNT_SECRETS_NAME'
SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH = 'SERVICE_ACCOUNT_CREDENTIAL_FILE_PATH'

# K8s constants
GCP_BACKEND_CONFIG_ANNOTATION = 'cloud.google.com/backend-config'
NODE_PORT_SERVICE_TYPE = 'NodePort'



from mage_ai.streaming.sources.kafka import KafkaSource
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class KafkaTests(TestCase):
    def test_init(self):
        with patch.object(KafkaSource, 'init_client') as mock_init_client:
            KafkaSource(dict(
                connector_type='kafka',
                bootstrap_server='test_server',
                consumer_group='test_group',
                topic='test_topic',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(KafkaSource, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                KafkaSource(dict(
                    connector_type='kafka',
                    bootstrap_server='test_server',
                    consumer_group='test_group',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'topic\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)

    def test_init_with_sasl_config(self):
        with patch.object(KafkaSource, 'init_client') as mock_init_client:
            source = KafkaSource(dict(
                connector_type='kafka',
                bootstrap_server='test_server',
                consumer_group='test_group',
                topic='test_topic',
                security_protocol='SASL_SSL',
                sasl_config=dict(
                    mechanism='PLAIN',
                    username='test_username',
                    password='test_password',
                ),
            ))
            mock_init_client.assert_called_once()
            self.assertEqual(source.config.security_protocol, 'SASL_SSL')
            self.assertEqual(source.config.sasl_config.mechanism, 'PLAIN')
            self.assertEqual(source.config.sasl_config.username, 'test_username')
            self.assertEqual(source.config.sasl_config.password, 'test_password')
            self.assertIsNone(source.config.ssl_config)

    def test_init_with_serde_config(self):
        with patch.object(KafkaSource, 'init_client') as mock_init_client:
            source = KafkaSource(dict(
                connector_type='kafka',
                bootstrap_server='test_server',
                consumer_group='test_group',
                topic='test_topic',
                serde_config=dict(
                    serialization_method='PROTOBUF',
                    schema_classpath='mage_ai.tests.base_test.TestCase',
                )
            ))
            mock_init_client.assert_called_once()
            self.assertEqual(source.config.serde_config.serialization_method, 'PROTOBUF')
            self.assertEqual(
                source.config.serde_config.schema_classpath, 'mage_ai.tests.base_test.TestCase')



from datetime import datetime
from mage_ai.data_cleaner.shared.utils import (
    is_geo_dataframe,
    is_spark_dataframe,
)
from mage_ai.data_preparation.models.variable import (
    Variable,
    VariableType,
    VARIABLE_DIR,
)
from mage_ai.data_preparation.storage.local_storage import LocalStorage
from mage_ai.data_preparation.repo_manager import (
    get_repo_config,
    get_repo_path,
    get_variables_dir,
)
from mage_ai.shared.constants import S3_PREFIX
from mage_ai.shared.dates import str_to_timedelta
from mage_ai.shared.utils import clean_name
from typing import Any, Dict, List
import os
import pandas as pd


class VariableManager:
    def __init__(self, repo_path=None, variables_dir=None):
        self.repo_path = repo_path or get_repo_path()
        if variables_dir is None:
            self.variables_dir = self.repo_path
        else:
            self.variables_dir = variables_dir
        self.storage = LocalStorage()
        # TODO: implement caching logic

    @classmethod
    def get_manager(
        self,
        repo_path: str = None,
        variables_dir: str = None,
    ) -> 'VariableManager':
        manager_args = dict(
            repo_path=repo_path,
            variables_dir=variables_dir,
        )
        if variables_dir is not None and variables_dir.startswith(S3_PREFIX):
            return S3VariableManager(**manager_args)
        else:
            return VariableManager(**manager_args)

    def add_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        data: Any,
        partition: str = None,
        variable_type: VariableType = None
    ) -> None:
        if type(data) is pd.DataFrame:
            variable_type = VariableType.DATAFRAME
        elif is_spark_dataframe(data):
            variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            variable_type = VariableType.GEO_DATAFRAME
        variable = Variable(
            clean_name(variable_uuid),
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            storage=self.storage,
            variable_type=variable_type,
        )
        # Delete data if it exists
        variable.delete()
        variable.variable_type = variable_type
        variable.write_data(data)

    async def add_variable_async(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        data: Any,
        partition: str = None,
        variable_type: VariableType = None
    ) -> None:
        if type(data) is pd.DataFrame:
            variable_type = VariableType.DATAFRAME
        elif is_spark_dataframe(data):
            variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            variable_type = VariableType.GEO_DATAFRAME
        variable = Variable(
            clean_name(variable_uuid),
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            storage=self.storage,
            variable_type=variable_type,
        )
        # Delete data if it exists
        variable.delete()
        variable.variable_type = variable_type
        await variable.write_data_async(data)

    def clean_variables(
        self,
        pipeline_uuid: str = None
    ):
        from mage_ai.data_preparation.models.pipeline import Pipeline

        repo_config = get_repo_config()
        if not repo_config.variables_retention_period:
            print('Variable retention period is not provided.')
            return
        min_partition = (datetime.utcnow() -
                         str_to_timedelta(repo_config.variables_retention_period)).strftime(
                            format='%Y%m%dT%H%M%S')
        print(f'Clean variables before partition {min_partition}')
        if pipeline_uuid is None:
            pipeline_uuids = Pipeline.get_all_pipelines(self.repo_path)
        else:
            pipeline_uuids = [pipeline_uuid]
        for pipeline_uuid in pipeline_uuids:
            print(f'Removing cached variables from pipeline {pipeline_uuid}')
            pipeline_variable_path = os.path.join(
                self.__pipeline_path(pipeline_uuid),
                VARIABLE_DIR,
            )
            dirs = self.storage.listdir(pipeline_variable_path)
            for dirname in dirs:
                if dirname.isdigit():
                    pipeline_schedule_vpath = os.path.join(pipeline_variable_path, dirname)
                    execution_partitions = self.storage.listdir(
                        pipeline_schedule_vpath,
                    )
                    for partition in execution_partitions:
                        if partition <= min_partition:
                            pipeline_partition_vpath = os.path.join(
                                pipeline_schedule_vpath,
                                partition,
                            )
                            print(f'Removing folder {pipeline_partition_vpath}')
                            self.storage.remove_dir(pipeline_partition_vpath)

    def delete_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        partition: str = None,
        variable_type: VariableType = None,
    ) -> None:
        Variable(
            variable_uuid,
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            storage=self.storage,
            variable_type=variable_type,
        ).delete()

    def get_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        dataframe_analysis_keys: List[str] = None,
        partition: str = None,
        variable_type: VariableType = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ) -> Any:
        variable = self.get_variable_object(
            pipeline_uuid,
            block_uuid,
            variable_uuid,
            partition=partition,
            variable_type=variable_type,
            spark=spark,
        )
        return variable.read_data(
            dataframe_analysis_keys=dataframe_analysis_keys,
            sample=sample,
            sample_count=sample_count,
            spark=spark,
        )

    def get_variable_object(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        partition: str = None,
        variable_type: VariableType = None,
        spark=None,
    ) -> Variable:
        if variable_type == VariableType.DATAFRAME and spark is not None:
            variable_type = VariableType.SPARK_DATAFRAME
        return Variable(
            variable_uuid,
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            spark=spark,
            storage=self.storage,
            variable_type=variable_type,
        )

    def get_variables_by_pipeline(self, pipeline_uuid: str) -> Dict[str, List[str]]:
        from mage_ai.data_preparation.models.pipeline import Pipeline
        pipeline = Pipeline.get(pipeline_uuid, repo_path=self.repo_path)
        variable_dir_path = os.path.join(self.__pipeline_path(pipeline_uuid), VARIABLE_DIR)
        if not self.storage.path_exists(variable_dir_path):
            return dict()
        block_dirs = self.storage.listdir(variable_dir_path)
        variables_by_block = dict()
        for d in block_dirs:
            if not pipeline.has_block(d) and d != 'global':
                continue
            block_variables_path = os.path.join(variable_dir_path, d)
            if not self.storage.isdir(block_variables_path):
                variables_by_block[d] = []
            else:
                variables = self.storage.listdir(os.path.join(variable_dir_path, d))
                variable_names = sorted([v.split('.')[0] for v in variables])
                variables_by_block[d] = [v for v in variable_names if v != '']
        return variables_by_block

    def get_variables_by_block(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        partition: str = None,
    ) -> List[str]:
        variable_dir_path = os.path.join(
            self.__pipeline_path(pipeline_uuid),
            VARIABLE_DIR,
            partition or '',
            clean_name(block_uuid),
        )
        if not self.storage.path_exists(variable_dir_path):
            return []
        variables = self.storage.listdir(variable_dir_path)
        return sorted([v.split('.')[0] for v in variables])

    def __pipeline_path(self, pipeline_uuid: str) -> str:
        path = os.path.join(self.variables_dir, 'pipelines', pipeline_uuid)
        if type(self.storage) is LocalStorage:
            if not self.storage.path_exists(path):
                self.storage.makedirs(path, exist_ok=True)
        return path


class S3VariableManager(VariableManager):
    def __init__(self, repo_path=None, variables_dir=None):
        super().__init__(repo_path=repo_path, variables_dir=variables_dir)
        from mage_ai.data_preparation.storage.s3_storage import S3Storage

        self.storage = S3Storage(dirpath=variables_dir)


def clean_variables(
    pipeline_uuid: str = None
):
    variables_dir = get_variables_dir()
    VariableManager(variables_dir=variables_dir).clean_variables(pipeline_uuid=pipeline_uuid)


def get_global_variables(
    pipeline_uuid: str,
) -> Dict[str, Any]:
    """
    Get all global variables. Global variables are stored together with project's code.
    """
    from mage_ai.data_preparation.models.pipeline import Pipeline
    pipeline = Pipeline.get(pipeline_uuid)
    if pipeline.variables is not None:
        global_variables = pipeline.variables
    else:
        variables_dir = get_variables_dir()
        variables = VariableManager(variables_dir=variables_dir).get_variables_by_block(
            pipeline_uuid,
            'global',
        )
        global_variables = dict()
        for variable in variables:
            global_variables[variable] = get_global_variable(pipeline_uuid, variable)

    return global_variables


def get_global_variable(
    pipeline_uuid: str,
    key: str,
) -> Any:
    """
    Get global variable by key. Global variables are stored together with project's code.
    """
    from mage_ai.data_preparation.models.pipeline import Pipeline
    pipeline = Pipeline.get(pipeline_uuid)
    if pipeline.variables is not None:
        return pipeline.variables.get(key)
    else:
        return VariableManager(variables_dir=get_variables_dir()).get_variable(
            pipeline_uuid,
            'global',
            key,
        )


def get_variable(
    pipeline_uuid: str,
    block_uuid: str,
    key: str,
    **kwargs
) -> Any:
    """
    Set block intermediate variable by key.
    Block intermediate variables are stored in variables dir.
    """
    return VariableManager(variables_dir=get_variables_dir()).get_variable(
        pipeline_uuid,
        block_uuid,
        key,
        **kwargs,
    )


def set_global_variable(
    pipeline_uuid: str,
    key: str,
    value: Any,
) -> None:
    """
    Set global variable by key. Global variables are stored together with project's code.
    """
    from mage_ai.data_preparation.models.pipeline import Pipeline
    pipeline = Pipeline.get(pipeline_uuid)

    if pipeline.variables is None:
        pipeline.variables = get_global_variables(pipeline_uuid)
    pipeline.update_global_variable(key, value)


def delete_global_variable(
    pipeline_uuid: str,
    key: str,
) -> None:
    """
    Delete global variable by key. Global variables are stored together with project's code.
    """
    from mage_ai.data_preparation.models.pipeline import Pipeline
    pipeline = Pipeline.get(pipeline_uuid)
    if pipeline.variables is not None:
        pipeline.delete_global_variable(key)
    else:
        VariableManager(
            variables_dir=get_variables_dir(),
        ).delete_variable(pipeline_uuid, 'global', key)



from datetime import datetime, timedelta
from mage_ai.services.hightouch.config import HightouchConfig
from mage_ai.services.hightouch.constants import (
    HIGHTOUCH_BASE_URL,
    DEFAULT_POLL_INTERVAL,
    PENDING_STATUSES,
    SUCCESS,
    TERMINAL_STATUSES,
)
from mage_ai.shared.http_client import HttpClient
from typing import Dict, Optional, Union
import time


class HightouchClient(HttpClient):
    BASE_URL = HIGHTOUCH_BASE_URL

    def __init__(self, config: Union[Dict, HightouchConfig]):
        if type(config) is dict:
            self.config = HightouchConfig.load(config=config)
        else:
            self.config = config
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': f'Bearer {self.config.api_key}',
        }

    def list_sources(self):
        return self.make_request('/sources')

    def get_source(self, source_id: int):
        return self.make_request(f'/sources/{source_id}')

    def list_destinations(self):
        return self.make_request('/destinations')

    def get_destination(self, destination_id: int):
        return self.make_request(f'/destinations/{destination_id}')

    def list_syncs(self):
        return self.make_request('/syncs')

    def get_sync(self, sync_id: int):
        return self.make_request(f'/syncs/{sync_id}')

    def list_sync_runs(self, sync_id: int, params: Dict = dict()):
        return self.make_request(f'/syncs/{sync_id}/runs', params=params)

    def trigger_sync(self, sync_id: int, payload: Dict = dict(fullResync=False)):
        return self.make_request(f'/syncs/{sync_id}/trigger', method='POST', payload=payload)

    def sync_and_poll(
        self,
        sync_id: int,
        payload: Dict = dict(fullResync=False),
        poll_interval: float = DEFAULT_POLL_INTERVAL,
        poll_timeout: Optional[float] = None,
    ):
        trigger_response = self.trigger_sync(sync_id, payload=payload)
        run_id = trigger_response.get('id')
        if not run_id:
            raise Exception(f'Failed to trigger Hightouch sync {sync_id}')
        poll_start = datetime.now()
        while True:
            sync_run = self.list_sync_runs(sync_id, dict(runId=run_id))['data'][0]
            print(
                f"Polling Hightouch Sync {sync_id}. Current status: {sync_run['status']}. "
                f"{100 * sync_run.get('completionRatio', 0)}% completed."
            )

            if sync_run['status'] in TERMINAL_STATUSES:
                print(f"Sync request status: {sync_run['status']}. Polling complete")
                if sync_run['error']:
                    print(f"Sync Request Error: {sync_run['error']}")

                if sync_run['status'] == SUCCESS:
                    break
                raise Exception(
                    f"Sync {sync_id} for request: {run_id} failed with status: "
                    f"{sync_run['error']} and error:  {sync_run['error']}"
                )
            if sync_run['status'] not in PENDING_STATUSES:
                print(
                    f"Unexpected status: {sync_run['status']} returned for sync {sync_id} and "
                    f"request {run_id}. Will try again, but if you see this error, please let "
                    "someone at Hightouch know.",
                )
            if (
                poll_timeout
                and datetime.now()
                > poll_start + timedelta(seconds=poll_timeout)
            ):
                raise Exception(
                    f"Sync {sync_id} for run: {run_id} time out after "
                    f"{datetime.now() - poll_start}. Last status was {sync_run['status']}."
                )
            time.sleep(poll_interval)



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from mage_ai.io.base import BaseSQLConnection, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.shared.hash import merge_dict
from pandas import DataFrame
from snowflake.connector import connect
from snowflake.connector.pandas_tools import write_pandas
from typing import Dict, List, Union


DEFAULT_LOGIN_TIMEOUT = 20
# NOTE: if credentials are wrong, it’ll take this many seconds for the user to be shown an error.
# TODO: check credentials before executing query and error sooner
# than waiting for Snowflake to timeout.
DEFAULT_NETWORK_TIMEOUT = 60 * 5


class Snowflake(BaseSQLConnection):
    """
    Handles data transfer between a Snowflake data warehouse and the Mage app.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to Snowflake data warehouse.
        The following arguments must be provided to the connector, all other
        arguments are optional.

        Required Arguments:
            user (str): Username for the Snowflake user.
            password (str): Login Password for the user.
            account (str): Snowflake account identifier (excluding
            `snowflake-computing.com` suffix).
        """
        if 'login_timeout' not in kwargs:
            kwargs['login_timeout'] = DEFAULT_LOGIN_TIMEOUT
        if 'network_timeout' not in kwargs:
            kwargs['network_timeout'] = DEFAULT_NETWORK_TIMEOUT
        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=kwargs.get('verbose', True), **kwargs)

    def default_database(self) -> str:
        return self.settings.get('database')

    def default_schema(self) -> str:
        return self.settings.get('schema')

    @property
    def timeout(self) -> int:
        return self.settings.get('timeout')

    def open(self) -> None:
        """
        Opens a connection to Snowflake.
        """
        with self.printer.print_msg('Connecting to Snowflake warehouse'):
            self._ctx = connect(**self.settings)

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Executes any query in the Snowflake data warehouse.

        Args:
            query_string (str): The query to execute on Snowflake's platform.
            **kwargs: Additional parameters to provide to the query
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                return cur.execute(query_string, timeout=self.timeout, **kwargs).fetchall()

    def execute_queries(
        self,
        queries: List[str],
        query_variables: List[Dict] = None,
        fetch_query_at_indexes: List[bool] = None,
        **kwargs,
    ) -> List:
        results = []

        with self.conn.cursor() as cursor:
            for idx, query in enumerate(queries):
                variables = query_variables[idx] \
                                if query_variables and idx < len(query_variables) \
                                else {}
                query = self._clean_query(query)

                if fetch_query_at_indexes and idx < len(fetch_query_at_indexes) and \
                        fetch_query_at_indexes[idx]:

                    rows = cursor.execute(query, timeout=self.timeout, **variables).fetchall()

                    full_table_name = fetch_query_at_indexes[idx]

                    columns = []
                    if type(full_table_name) is str:
                        columns = self.get_columns(
                            cursor,
                            full_table_name=full_table_name,
                        )
                    elif len(rows) >= 1 and len(rows[0]) >= 1:
                        columns = [f'col_{i}' for i in range(len(rows[0]))]

                    result = DataFrame(rows, columns=columns)
                else:
                    result = cursor.execute(query, timeout=self.timeout, **variables)

                results.append(result)

        return results

    def get_columns(
        self,
        cursor,
        database: str = None,
        schema: str = None,
        table_name: str = None,
        full_table_name: str = None,
    ) -> List[str]:
        columns = None
        if not full_table_name and database and schema and table_name:
            full_table_name = f'"{database}"."{schema}"."{table_name}"'

        if full_table_name:
            arr = cursor.execute(
                f'DESCRIBE TABLE {full_table_name}',
                timeout=self.timeout,
            ).fetchall()
            columns = [t[0] for t in arr]

        return columns

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        database: str = None,
        schema: str = None,
        table_name: str = None,
        full_table_name: str = None,
        *args,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from Snowflake into a Pandas data frame based on the query given.
        This will fail unless a `SELECT` query is provided. This function will load at
        maximum 10,000,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.

        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults
                                    to 10,000,000.
            *args, **kwargs: Additional parameters to provide to the query

        Returns:
            DataFrame: Data frame associated with the given query.
        """
        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n{query_string}'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            with self.conn.cursor() as cur:
                columns = None
                if (database and schema and table_name) or full_table_name:
                    columns = self.get_columns(
                        cur,
                        database=database,
                        schema=schema,
                        table_name=table_name,
                        full_table_name=full_table_name,
                    )

                results = cur.execute(
                    self._enforce_limit(query_string, limit),
                    *args,
                    timeout=self.timeout,
                    **kwargs,
                ).fetchall()

                if not columns and len(results) >= 1:
                    columns = [f'col{i}' for i in range(len(results[0]))]

                return DataFrame(results, columns=columns)

    def export(
        self,
        df: DataFrame,
        table_name: str,
        database: str,
        schema: str,
        if_exists: str = 'append',
        query_string: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports a Pandas data frame to a Snowflake warehouse based on the table name.
        If table doesn't exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export to a Snowflake warehouse.
            table_name (str): Name of the table to export data to (excluding database and schema).
            database (str): Name of the database in which the table is located.
            schema (str): Name of the schema in which the table is located.
            if_exists (str, optional): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table.
            Defaults to `'append'`.
            **kwargs: Additional arguments to pass to writer
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        def __process():
            with self._ctx.cursor() as cur:
                cur.execute(
                    f'CREATE SCHEMA IF NOT EXISTS {database}.{schema}',
                    timeout=self.timeout,
                )

                cur.execute(f'USE DATABASE {database}', timeout=self.timeout)
                cur.execute(
                    'SELECT * FROM information_schema.tables WHERE table_schema = ' +
                    f'\'{schema}\' AND table_name = \'{table_name}\'',
                    timeout=self.timeout,
                )

                table_exists = cur.rowcount >= 1
                should_create_table = not table_exists

                if table_exists:
                    if cur.rowcount > 1:
                        raise ValueError(
                            f'Two or more tables with the name {table_name} are found.',
                        )

                    if ExportWritePolicy.FAIL == if_exists:
                        raise RuntimeError(
                            f'Table {table_name} already exists in the current warehouse, '
                            'database, schema scenario.'
                        )
                    elif ExportWritePolicy.REPLACE == if_exists:
                        cur.execute(f'USE DATABASE {database}', timeout=self.timeout)
                        cur.execute(f'DROP TABLE "{schema}"."{table_name}"', timeout=self.timeout)
                        should_create_table = True

                if query_string:
                    cur.execute(f'USE DATABASE {database}', timeout=self.timeout)

                    if should_create_table:
                        cur.execute(f"""
CREATE TABLE IF NOT EXISTS "{database}"."{schema}"."{table_name}" AS
{query_string}
""", timeout=self.timeout)
                    else:
                        cur.execute(f"""
INSERT INTO "{database}"."{schema}"."{table_name}"
{query_string}
""", timeout=self.timeout)

                else:
                    write_pandas(
                        self.conn,
                        df,
                        table_name,
                        database=database,
                        schema=schema,
                        auto_create_table=should_create_table,
                        **kwargs,
                    )

        if verbose:
            with self.printer.print_msg(
                f'Exporting data to \'{database}.{schema}.{table_name}\''
            ):
                __process()
        else:
            __process()

    @classmethod
    def with_config(
        cls,
        config: BaseConfigLoader,
        database=None,
        schema=None,
        **kwargs,
    ) -> 'Snowflake':
        """
        Initializes Snowflake client from configuration loader.

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        conn_kwargs = dict(
            account=config[ConfigKey.SNOWFLAKE_ACCOUNT],
            database=database or config[ConfigKey.SNOWFLAKE_DEFAULT_DB],
            schema=schema or config[ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA],
            user=config[ConfigKey.SNOWFLAKE_USER],
            warehouse=config[ConfigKey.SNOWFLAKE_DEFAULT_WH],
        )

        if ConfigKey.SNOWFLAKE_TIMEOUT in config:
            conn_kwargs['timeout'] = config[ConfigKey.SNOWFLAKE_TIMEOUT]

        if ConfigKey.SNOWFLAKE_ROLE in config:
            conn_kwargs['role'] = config[ConfigKey.SNOWFLAKE_ROLE]

        if ConfigKey.SNOWFLAKE_PASSWORD in config:
            conn_kwargs['password'] = config[ConfigKey.SNOWFLAKE_PASSWORD]
        elif ConfigKey.SNOWFLAKE_PRIVATE_KEY_PATH in config:
            with open(config[ConfigKey.SNOWFLAKE_PRIVATE_KEY_PATH], 'rb') as key:
                password = None
                if ConfigKey.SNOWFLAKE_PRIVATE_KEY_PASSPHRASE in config:
                    password = config[ConfigKey.SNOWFLAKE_PRIVATE_KEY_PASSPHRASE].encode()

                p_key = serialization.load_pem_private_key(
                    key.read(),
                    password=password,
                    backend=default_backend()
                )

            pkb = p_key.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            conn_kwargs['private_key'] = pkb

        return cls(
            **merge_dict(conn_kwargs, kwargs),
        )



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from pymongo import MongoClient
from typing import Dict, List
import time


@dataclass
class MongoDbConfig(BaseConfig):
    connection_string: str
    database_name: str
    collection_name: str


class MongoDbSink(BaseSink):
    config_class = MongoDbConfig

    def init_client(self):
        self.client = MongoClient(self.config.connection_string)
        self.database = self.client[self.config.database_name]
        self.collection = self.database[self.config.collection_name]

    def write(self, data: Dict):
        if type(data) is dict:
            self.collection.insert_one(data)
        else:
            self.collection.insert_one({"data": data})
        self._print(f'[MongoDB] Ingest data {data}, time={time.time()}')

    def batch_write(self, data: List[Dict]):
        if not data:
            return
        output_docs = []
        for doc in data:
            if type(doc) is dict:
                output_docs.append(doc)
            else:
                output_docs.append({"data": doc})
        self.collection.insert_many(output_docs)
        self._print(f'[MongoDB] Batch ingest {len(data)} records, time={time.time()}.')



from concurrent.futures import ThreadPoolExecutor
from joblib import Parallel, delayed
from threading import Thread

MAX_WORKERS = 16


def execute_parallel(list_of_funcs_and_args, verbose=0):
    parallel = Parallel(n_jobs=-1, prefer='threads', verbose=verbose)
    return parallel(delayed(func)(*args) for func, args in list_of_funcs_and_args)


def start_thread(target, **kwargs):
    thread = Thread(
        target=target,
        kwargs=kwargs,
    )
    thread.start()

    return thread


def parallelize(func, arr):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as pool:
        return pool.map(func, arr)


def parallelize_multiple_args(func, arr_args):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as pool:
        return pool.map(func, *zip(*arr_args))


def run_parallel_threads(list_of_funcs_and_args_or_kwargs):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as pool:
        for func, args in list_of_funcs_and_args_or_kwargs:
            pool.submit(func, *args)


def run_parallel(func, arr_args_1, arr_args_2):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as pool:
        return pool.map(func, *arr_args_1, *arr_args_2)


def run_parallel_multiple_args(func, *args):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as pool:
        return pool.map(func, *args)



from mage_ai.shared.array import find, unique_by
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.utils import clean_name as clean_name_orig
from typing import Any, Dict, List, Tuple
import json
import pandas as pd


def clean_name(name: str, **kwargs) -> str:
    return clean_name_orig(name, allow_characters=['/'], **kwargs)


def dynamic_block_uuid(
    block_uuid: str,
    metadata: Dict,
    index: int,
    upstream_block_uuid: str = None,
) -> str:
    block_uuid_subname = metadata.get('block_uuid', index)
    uuid = f'{block_uuid}:{block_uuid_subname}'

    if upstream_block_uuid:
        parts = upstream_block_uuid.split(':')
        if len(parts) >= 2:
            upstream_indexes = ':'.join(parts[1:])
            uuid = f'{uuid}:{upstream_indexes}'

    return uuid


def create_block_run_from_dynamic_child(
    block,
    pipeline_run,
    block_metadata: Dict,
    index: int,
    upstream_block_uuid: str = None,
):
    metadata = block_metadata.copy()
    metadata.update(dict(dynamic_block_index=index))

    block_uuid = dynamic_block_uuid(
        block.uuid,
        metadata,
        index,
        upstream_block_uuid=upstream_block_uuid,
    )
    block_run = pipeline_run.create_block_run(block_uuid, metrics=metadata)

    return block_run


def dynamic_block_values_and_metadata(
    block,
    execution_partition: str = None,
    block_uuid: str = None,
):
    block_uuid_original = block.uuid
    block_uuid = block_uuid_original if block_uuid is None else block_uuid

    values = []
    block_metadata = []
    output_vars = block.output_variables(execution_partition=execution_partition)
    for idx, output_name in enumerate(output_vars):
        if idx == 0:
            values = block.pipeline.variable_manager.get_variable(
                block.pipeline.uuid,
                block_uuid,
                output_name,
                partition=execution_partition,
            )
        elif idx == 1:
            block_metadata = block.pipeline.variable_manager.get_variable(
                block.pipeline.uuid,
                block_uuid,
                output_name,
                partition=execution_partition,
            )

    return values, block_metadata


def create_block_runs_from_dynamic_block(
    block,
    pipeline_run,
    block_uuid: str = None,
) -> List:
    block_uuid_original = block.uuid
    block_uuid = block_uuid_original if block_uuid is None else block_uuid
    execution_partition = pipeline_run.execution_partition

    values, block_metadata = dynamic_block_values_and_metadata(
        block,
        execution_partition,
        block_uuid,
    )

    all_block_runs = []
    # Dynamic child blocks (aka created from a dynamic block)
    for downstream_block in block.downstream_blocks:
        is_dynamic = is_dynamic_block(downstream_block)
        should_reduce = should_reduce_output(downstream_block)
        descendants = get_all_descendants(downstream_block)

        block_runs_created_by_block_uuid = {}
        dynamic_child_block_runs = []
        for idx, value in enumerate(values):
            if idx < len(block_metadata):
                metadata = block_metadata[idx].copy()
            else:
                metadata = {}

            arr = []
            for upstream_block in downstream_block.upstream_blocks:
                if block_uuid_original == upstream_block.uuid and block_uuid_original != block_uuid:
                    arr.append(block_uuid)
                else:
                    arr.append(upstream_block.uuid)

            block_run = create_block_run_from_dynamic_child(
                downstream_block,
                pipeline_run,
                merge_dict(metadata, dict(
                    dynamic_upstream_block_uuids=arr,
                )),
                idx,
                upstream_block_uuid=block_uuid,
            )
            all_block_runs.append(block_run)
            dynamic_child_block_runs.append(block_run)

            if is_dynamic or should_reduce:
                continue

            # Schedule all descendants
            for b in descendants:
                # If block has dynamic upstream, skip since creation of downstream
                # is handled in pipeline scheduler
                if find(lambda x: is_dynamic_block(x), b.upstream_blocks):
                    continue

                """
                If a descendant has an immediate upstream block that is different than
                the current downstream_block that reduces output, skip this loop because
                this descendant will be created by that other upstream block that reduces output
                """
                # if find(
                #     lambda x: downstream_block.uuid != x.uuid and should_reduce_output(x),
                #     b.upstream_blocks,
                # ):
                #     continue

                arr = []
                for upstream_block in b.upstream_blocks:
                    ancestors = get_all_ancestors(upstream_block)
                    # If the upstream block has the current dynamic child as an ancestor,
                    # then have this block depend on a block UUID with the dynamic UUID suffix;
                    # e.g. block_uuid:index
                    if downstream_block.uuid in [a.uuid for a in ancestors]:
                        arr.append(dynamic_block_uuid(
                            upstream_block.uuid,
                            metadata,
                            idx,
                        ))
                    elif downstream_block.uuid == upstream_block.uuid:
                        arr.append(block_run.block_uuid)
                    elif is_dynamic_block_child(upstream_block):
                        # Needs to know that the ancestors are dynamic
                        # or dynamic child without reduce
                        arr.append(dynamic_block_uuid(
                            upstream_block.uuid,
                            metadata,
                            idx,
                        ))
                    else:
                        arr.append(upstream_block.uuid)

                b_uuid = dynamic_block_uuid(b.uuid, metadata, idx)
                if b_uuid in block_runs_created_by_block_uuid:
                    continue
                block_runs_created_by_block_uuid[b_uuid] = True

                br = create_block_run_from_dynamic_child(
                    b,
                    pipeline_run,
                    merge_dict(metadata, dict(
                        dynamic_upstream_block_uuids=arr,
                    )),
                    idx,
                )
                all_block_runs.append(br)

        if should_reduce:
            for b in descendants:
                ancestors = get_all_ancestors(b)
                unique_dynamic_ancestors = unique_by(
                    list(filter(lambda x: is_dynamic_block(x), ancestors)),
                    lambda x: x.uuid,
                )

                skip_creating_downstream = False
                ancestors_uuids = [a.uuid for a in ancestors]

                for dynamic_ancestor in unique_dynamic_ancestors:
                    if skip_creating_downstream:
                        break

                    """
                    If any dynamic ancestors doesn't have reduce block as the descendants, skip
                    creating this downstream block.
                    """
                    dynamic_ancestor_descendants = get_all_descendants(dynamic_ancestor)

                    dynamic_ancestor_has_reduce_block = False
                    for d in dynamic_ancestor_descendants:
                        if d.uuid in ancestors_uuids:
                            if should_reduce_output(d):
                                dynamic_ancestor_has_reduce_block = True
                                break
                    if not dynamic_ancestor_has_reduce_block:
                        skip_creating_downstream = True
                        break

                    down_uuids_as_ancestors = []
                    for down in dynamic_ancestor.downstream_blocks:
                        if down.uuid in ancestors_uuids and not should_reduce_output(down):
                            down_uuids_as_ancestors.append(down.uuid)
                    skip_creating_downstream = len(down_uuids_as_ancestors) >= 2

                # Only create downstream block runs if it doesn’t have dynamically created upstream
                # blocks (aka dynamic child) that were created by a 2nd ancestor that is a dynamic
                # block
                if skip_creating_downstream:
                    continue

                for upstream_block in b.upstream_blocks:
                    if should_reduce_output(upstream_block) and \
                            upstream_block.uuid != downstream_block.uuid:
                        skip_creating_downstream = True
                        break
                """
                If an upstream block needs to be reduced but it's not the same with the current
                block that's being reduced, skip the downstream block creation.
                """
                if skip_creating_downstream:
                    continue

                arr = []
                for upstream_block in b.upstream_blocks:
                    if downstream_block.uuid == upstream_block.uuid:
                        arr += [b.block_uuid for b in dynamic_child_block_runs]
                    else:
                        arr.append(upstream_block.uuid)

                all_block_runs.append(pipeline_run.create_block_run(
                    b.uuid,
                    metrics=dict(dynamic_upstream_block_uuids=arr),
                    skip_if_exists=True,
                ))

    return all_block_runs


def get_all_ancestors(block) -> List:
    arr = get_leaf_nodes(block, 'upstream_blocks', include_all_nodes=True)
    return list(filter(
        lambda x: x.uuid != block.uuid,
        arr,
    ))


def get_all_descendants(block) -> List:
    arr = get_leaf_nodes(block, 'downstream_blocks', include_all_nodes=True)
    return list(filter(
        lambda x: x.uuid != block.uuid,
        arr,
    ))


def get_leaf_nodes(
    block,
    attribute_key: str,
    condition=None,
    include_all_nodes: bool = False,
) -> List:
    leafs = []

    def _get_leaf_nodes(b):
        if condition is None or condition(b):
            if b is not None:
                arr = getattr(b, attribute_key)
                if len(arr) == 0 or (include_all_nodes and b.uuid != block.uuid):
                    leafs.append(b)

                for n in arr:
                    _get_leaf_nodes(n)

    _get_leaf_nodes(block)

    return leafs


def is_dynamic_block(block) -> bool:
    return block.configuration and block.configuration.get('dynamic', False)


def should_reduce_output(block) -> bool:
    return block.configuration and block.configuration.get('reduce_output', False)


def is_dynamic_block_child(block) -> bool:
    dynamic_or_child = []

    for upstream_block in block.upstream_blocks:
        if is_dynamic_block(upstream_block) or is_dynamic_block_child(upstream_block):
            dynamic_or_child.append(upstream_block)

    if len(dynamic_or_child) == 0:
        return False

    dynamic_or_child_with_reduce = list(filter(lambda x: should_reduce_output(x), dynamic_or_child))

    return len(block.upstream_blocks) > len(dynamic_or_child_with_reduce)


def output_variables(
    pipeline,
    block_uuid: str,
    execution_partition: str = None,
) -> List[str]:
    """Return output variables in dictionary.
    The key is the variable name, and the value is variable data type.

    Args:
        execution_partition (str, optional): The execution paratition string.

    Returns:
        List[str]: List of variable names.
    """
    all_variables = pipeline.variable_manager.get_variables_by_block(
        pipeline.uuid,
        block_uuid,
        partition=execution_partition,
    )
    output_variables = [v for v in all_variables if is_output_variable(v)]
    output_variables.sort()

    return output_variables


def is_output_variable(variable_uuid: str) -> bool:
    return variable_uuid == 'df' or variable_uuid.startswith('output')


def is_valid_print_variable(k, v, block_uuid):
    if f'{block_uuid}_' not in k:
        return False
    if type(v) is not str:
        return False
    # Not store empty json
    if v == '{}' or v == '':
        return False
    try:
        json_data = json.loads(v)
        # Not store empty data in json object
        if 'data' in json_data and not json_data['data']:
            return False
    except Exception:
        pass
    return True


def input_variables(
    pipeline,
    upstream_block_uuids: List[str],
    execution_partition: str = None,
) -> Dict[str, List[str]]:
    """Get input variables from upstream blocks' output variables.
    Args:
        execution_partition (str, optional): The execution paratition string.

    Returns:
        Dict[str, List[str]]: Mapping from upstream block uuid to a list of variable names
    """
    return {block_uuid: output_variables(pipeline, block_uuid, execution_partition)
            for block_uuid in upstream_block_uuids}


def fetch_input_variables(
    pipeline,
    upstream_block_uuids: List[str],
    input_args: List[Any],
    execution_partition: str = None,
    global_vars: Dict = None,
    dynamic_block_index: int = None,
    dynamic_upstream_block_uuids: List[str] = None,
) -> Tuple[List, List, List]:
    spark = (global_vars or dict()).get('spark')
    upstream_block_uuids_final = []

    kwargs_vars = []

    input_vars = []
    if input_args is not None:
        input_vars = input_args
    elif pipeline is not None:
        input_vars = [None for i in range(len(upstream_block_uuids))]
        input_variables_by_uuid = input_variables(
            pipeline,
            upstream_block_uuids,
            execution_partition,
        )
        keys = input_variables_by_uuid.keys()
        reduce_output_indexes = []

        for idx, upstream_block_uuid in enumerate(keys):
            upstream_block_uuids_final.append(upstream_block_uuid)
            upstream_block = pipeline.get_block(upstream_block_uuid)
            should_reduce = should_reduce_output(upstream_block)

            variables = input_variables_by_uuid[upstream_block_uuid]
            variable_values = [
                pipeline.variable_manager.get_variable(
                    pipeline.uuid,
                    upstream_block_uuid,
                    var,
                    partition=execution_partition,
                    spark=spark,
                )
                for var in variables
            ]

            upstream_in_dynamic_upstream = dynamic_upstream_block_uuids and find(
                lambda x: upstream_block_uuid in x,
                dynamic_upstream_block_uuids or [],
            )

            if dynamic_upstream_block_uuids and (should_reduce or upstream_in_dynamic_upstream):
                reduce_output_indexes.append((idx, upstream_block_uuid))
            elif is_dynamic_block(upstream_block):
                val = None
                if len(variable_values) >= 1:
                    arr = variable_values[0]
                    index_to_use = 0 if dynamic_block_index is None else dynamic_block_index
                    if type(arr) is list and len(arr) >= 1 and index_to_use < len(arr):
                        val = arr[index_to_use]
                input_vars[idx] = val

                # output_0 is the metadata for dynamic blocks
                if len(variable_values) >= 2:
                    arr = variable_values[1]
                    index_to_use = 0 if dynamic_block_index is None else dynamic_block_index
                    if type(arr) is list and len(arr) >= 1 and index_to_use < len(arr):
                        val = arr[index_to_use]
                    kwargs_vars.append(val)
            elif not dynamic_upstream_block_uuids or not upstream_in_dynamic_upstream:
                if type(variable_values) is list and len(variable_values) == 1:
                    final_val = variable_values[0]
                else:
                    final_val = variable_values

                if should_reduce:
                    final_val = [final_val]

                input_vars[idx] = final_val

        if dynamic_upstream_block_uuids:
            for tup in reduce_output_indexes:
                idx = tup[0]
                upstream_block_uuid = tup[1]

                upstream_block = pipeline.get_block(upstream_block_uuid)
                should_reduce = should_reduce_output(upstream_block)
                upstream_is_dynamic = is_dynamic_block(upstream_block)

                uuids = list(
                    filter(lambda x: upstream_block_uuid in x, dynamic_upstream_block_uuids),
                )
                input_variables_by_uuid = input_variables(
                    pipeline,
                    uuids,
                    execution_partition,
                )

                final_value = []
                for upstream_block_uuid, variables in input_variables_by_uuid.items():
                    end_idx = 1 if upstream_is_dynamic else len(variables)
                    for var in variables[:end_idx]:
                        variable_values = pipeline.variable_manager.get_variable(
                            pipeline.uuid,
                            upstream_block_uuid,
                            var,
                            partition=execution_partition,
                            spark=spark,
                        )

                        if type(variable_values) is list and len(variable_values) == 1:
                            val = variable_values[0]
                        else:
                            val = variable_values

                        if type(val) is list:
                            final_value += val
                        else:
                            final_value.append(val)

                    # output_0 is the metadata for dynamic blocks
                    if dynamic_block_index is not None and \
                            upstream_is_dynamic and \
                            len(variables) >= 2:

                        var = variables[1]
                        variable_values = pipeline.variable_manager.get_variable(
                            pipeline.uuid,
                            upstream_block_uuid,
                            var,
                            partition=execution_partition,
                            spark=spark,
                        )
                        if dynamic_block_index < len(variable_values):
                            val = variable_values[dynamic_block_index]
                            kwargs_vars.append(val)

                if len(final_value) >= 1 and all([type(v) is pd.DataFrame for v in final_value]):
                    final_value = pd.concat(final_value)

                if not should_reduce:
                    # Only get the 1st output of a dynamic block;
                    # the 2nd output is the dynamic child block’s metadata
                    if upstream_is_dynamic and dynamic_block_index is not None:
                        final_value = final_value[dynamic_block_index]

                    if type(final_value) is not pd.DataFrame and \
                        type(final_value) is list and \
                            len(final_value) == 1:
                        final_value = final_value[0]
                input_vars[idx] = final_value

    return input_vars, kwargs_vars, upstream_block_uuids_final



"""Add pipeline_uuid index to pipeline_schedule.

Revision ID: 21e31d66ccea
Revises: 1a737bdee53b
Create Date: 2023-04-30 17:35:47.150966

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '21e31d66ccea'
down_revision = '1a737bdee53b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('pipeline_schedule', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_pipeline_schedule_pipeline_uuid'), ['pipeline_uuid'], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('pipeline_schedule', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_pipeline_schedule_pipeline_uuid'))

    # ### end Alembic commands ###



from dataclasses import dataclass, field
from datetime import datetime
from mage_ai.data_preparation.models.constants import (
    PIPELINES_FOLDER,
)
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.shared.config import BaseConfig
from typing import Dict, List
import enum
import os
import traceback
import yaml

TRIGGER_FILE_NAME = 'triggers.yaml'


class ScheduleStatus(str, enum.Enum):
    ACTIVE = 'active'
    INACTIVE = 'inactive'


class ScheduleType(str, enum.Enum):
    API = 'api'
    EVENT = 'event'
    TIME = 'time'


class ScheduleInterval(str, enum.Enum):
    ONCE = '@once'
    HOURLY = '@hourly'
    DAILY = '@daily'
    WEEKLY = '@weekly'
    MONTHLY = '@monthly'


@dataclass
class SettingsConfig(BaseConfig):
    skip_if_previous_running: bool = False
    allow_blocks_to_fail: bool = False


@dataclass
class Trigger(BaseConfig):
    name: str
    pipeline_uuid: str
    schedule_type: ScheduleType
    start_time: datetime
    schedule_interval: str
    status: ScheduleStatus = ScheduleStatus.INACTIVE
    variables: Dict = field(default_factory=dict)
    sla: int = None     # in seconds
    settings: Dict = field(default_factory=dict)

    def __post_init__(self):
        if self.schedule_type and type(self.schedule_type) is str:
            self.schedule_type = ScheduleType(self.schedule_type)
        if self.status and type(self.status) is str:
            self.status = ScheduleStatus(self.status)


def get_triggers_by_pipeline(pipeline_uuid: str) -> List[Trigger]:
    pipeline_path = os.path.join(get_repo_path(), PIPELINES_FOLDER, pipeline_uuid)
    trigger_file_path = os.path.join(pipeline_path, TRIGGER_FILE_NAME)
    if not os.path.exists(trigger_file_path):
        return []
    try:
        with open(trigger_file_path) as fp:
            content = fp.read()
        triggers = load_trigger_configs(content, pipeline_uuid=pipeline_uuid)
    except Exception:
        traceback.print_exc()
        triggers = []

    return triggers


def load_trigger_configs(
    content: str,
    pipeline_uuid: str = None,
    raise_exception: bool = False,
) -> List[Trigger]:
    yaml_config = yaml.safe_load(content) or {}
    trigger_configs = yaml_config.get('triggers') or {}

    triggers = []
    for trigger_config in trigger_configs:
        if pipeline_uuid:
            trigger_config['pipeline_uuid'] = pipeline_uuid
        try:
            trigger = Trigger.load(config=trigger_config)
            triggers.append(trigger)
        except Exception as e:
            if raise_exception:
                raise Exception(
                    f'Failed to parse trigger config {trigger_config}. {str(e)}') from e
            else:
                traceback.print_exc()
    return triggers
